# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later


import binascii
import contextlib
import os
import os.path
from pathlib import Path
import tempfile
import unittest
import unittest.mock

from _drgn_util.elf import ET, PT, SHF, SHT
from drgn import (
    DebugInfoOptions,
    MainModule,
    MissingDebugInfoError,
    ModuleFileStatus,
    Program,
    SharedLibraryModule,
    SupplementaryFileKind,
    VdsoModule,
)
from tests import TestCase
from tests.dwarfwriter import create_dwarf_file
from tests.elfwriter import ElfSection, create_elf_file
from tests.resources import get_resource

ALLOCATED_SECTION = ElfSection(
    name=".bss",
    sh_type=SHT.PROGBITS,
    sh_flags=SHF.ALLOC,
    p_type=PT.LOAD,
    vaddr=0x10000000,
    memsz=0x1000,
)


@contextlib.contextmanager
def NamedTemporaryElfFile(*, loadable=True, debug=True, sections=(), **kwargs):
    if loadable:
        sections = (ALLOCATED_SECTION,) + sections
    with tempfile.NamedTemporaryFile() as f:
        if debug:
            f.write(create_dwarf_file((), sections=sections, **kwargs))
        else:
            f.write(create_elf_file(ET.EXEC, sections=sections, **kwargs))
        f.flush()
        yield f


class TestModuleTryFile(TestCase):
    def setUp(self):
        self.prog = Program()
        self.prog.set_enabled_debug_info_finders([])

    def test_want_both(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile() as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)

        for status in set(ModuleFileStatus) - {ModuleFileStatus.HAVE}:
            for file in ("loaded", "debug"):
                with self.subTest(file=file):
                    self.assertEqual(getattr(module, f"wants_{file}_file")(), False)
                    # Test that we can't unset the file once it's set.
                    status_attr = file + "_file_status"
                    with self.subTest(from_=ModuleFileStatus.HAVE, to=status):
                        self.assertRaises(
                            ValueError, setattr, module, status_attr, status
                        )
                        self.assertEqual(
                            getattr(module, status_attr), ModuleFileStatus.HAVE
                        )

    def test_want_both_not_loadable(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile(loadable=False) as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)

    def test_want_both_no_debug(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile(debug=False) as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.debug_file_path)

    def test_want_both_is_neither(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile(loadable=False, debug=False) as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertIsNone(module.debug_file_path)

    def test_only_want_loaded(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.debug_file_status = ModuleFileStatus.DONT_WANT
        with NamedTemporaryElfFile() as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.DONT_WANT)
        self.assertIsNone(module.debug_file_path)

    def test_only_want_loaded_not_loadable(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.debug_file_status = ModuleFileStatus.DONT_WANT
        with NamedTemporaryElfFile(loadable=False) as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.DONT_WANT)
        self.assertIsNone(module.debug_file_path)

    def test_only_want_loaded_no_debug(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.debug_file_status = ModuleFileStatus.DONT_WANT
        with NamedTemporaryElfFile(debug=False) as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.DONT_WANT)
        self.assertIsNone(module.debug_file_path)

    def test_only_want_loaded_is_neither(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.debug_file_status = ModuleFileStatus.DONT_WANT
        with NamedTemporaryElfFile(loadable=False, debug=False) as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.DONT_WANT)
        self.assertIsNone(module.debug_file_path)

    def test_only_want_debug(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.loaded_file_status = ModuleFileStatus.DONT_WANT
        with NamedTemporaryElfFile() as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.DONT_WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)

    def test_only_want_debug_not_loadable(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.loaded_file_status = ModuleFileStatus.DONT_WANT
        with NamedTemporaryElfFile(loadable=False) as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.DONT_WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)

    def test_only_want_debug_no_debug(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.loaded_file_status = ModuleFileStatus.DONT_WANT
        with NamedTemporaryElfFile(debug=False) as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.DONT_WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.debug_file_path)

    def test_only_want_debug_is_neither(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.loaded_file_status = ModuleFileStatus.DONT_WANT
        with NamedTemporaryElfFile(loadable=False, debug=False) as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.DONT_WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.debug_file_path)

    def test_want_neither(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.loaded_file_status = ModuleFileStatus.DONT_WANT
        module.debug_file_status = ModuleFileStatus.DONT_WANT
        with NamedTemporaryElfFile() as f:
            module.try_file(f.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.DONT_WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.DONT_WANT)
        self.assertIsNone(module.debug_file_path)

    def test_separate_files_loaded_first(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile(debug=False) as f1:
            module.try_file(f1.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f1.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.debug_file_path)

        with NamedTemporaryElfFile(loadable=False) as f2:
            module.try_file(f2.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f1.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f2.name)

    def test_separate_files_debug_first(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile(loadable=False) as f1:
            module.try_file(f1.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f1.name)

        with NamedTemporaryElfFile(debug=False) as f2:
            module.try_file(f2.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f2.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f1.name)

    def test_loadable_then_both(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile(debug=False) as f1:
            module.try_file(f1.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f1.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.debug_file_path)

        with NamedTemporaryElfFile() as f2:
            module.try_file(f2.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f1.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f2.name)

    def test_debug_then_both(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile(loadable=False) as f1:
            module.try_file(f1.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f1.name)

        with NamedTemporaryElfFile() as f2:
            module.try_file(f2.name)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f2.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f1.name)

    def test_no_build_id_force(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile() as f:
            module.try_file(f.name, force=True)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)

    def test_no_build_id_file_has_build_id(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile(build_id=b"\x01\x23\x45\x67\x89\xab\xcd\xef") as f:
            module.try_file(f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)
        self.assertEqual(module.build_id, b"\x01\x23\x45\x67\x89\xab\xcd\xef")

    def test_no_build_id_file_has_build_id_force(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile(build_id=b"\x01\x23\x45\x67\x89\xab\xcd\xef") as f:
            module.try_file(f.name, force=True)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)
        self.assertEqual(module.build_id, b"\x01\x23\x45\x67\x89\xab\xcd\xef")

    def test_build_id_match(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        with NamedTemporaryElfFile(build_id=b"\x01\x23\x45\x67\x89\xab\xcd\xef") as f:
            module.try_file(f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)
        self.assertEqual(module.build_id, b"\x01\x23\x45\x67\x89\xab\xcd\xef")

    def test_build_id_match_force(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        with NamedTemporaryElfFile(build_id=b"\x01\x23\x45\x67\x89\xab\xcd\xef") as f:
            module.try_file(f.name, force=True)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)
        self.assertEqual(module.build_id, b"\x01\x23\x45\x67\x89\xab\xcd\xef")

    def test_build_id_mismatch(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        with NamedTemporaryElfFile(build_id=b"\xff\xff\xff\xff") as f:
            module.try_file(f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.debug_file_path)
        self.assertEqual(module.build_id, b"\x01\x23\x45\x67\x89\xab\xcd\xef")

    def test_build_id_mismatch_force(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        with NamedTemporaryElfFile(build_id=b"\xff\xff\xff\xff") as f:
            module.try_file(f.name, force=True)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)
        self.assertEqual(module.build_id, b"\x01\x23\x45\x67\x89\xab\xcd\xef")

    def test_build_id_missing(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        with NamedTemporaryElfFile() as f:
            module.try_file(f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.debug_file_path)
        self.assertEqual(module.build_id, b"\x01\x23\x45\x67\x89\xab\xcd\xef")

    def test_build_id_missing_force(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        with NamedTemporaryElfFile() as f:
            module.try_file(f.name, force=True)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)
        self.assertEqual(module.build_id, b"\x01\x23\x45\x67\x89\xab\xcd\xef")

    def test_gnu_debugaltlink(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            alt_path = debug_dir / "alt.debug"
            alt_path.write_bytes(create_dwarf_file((), build_id=alt_build_id))

            binary_path = bin_dir / "binary"
            binary_path.write_bytes(
                create_dwarf_file(
                    (),
                    sections=(ALLOCATED_SECTION,),
                    build_id=build_id,
                    gnu_debugaltlink=(alt_path, alt_build_id),
                )
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            module.build_id = build_id

            self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
            self.assertRaises(ValueError, module.wanted_supplementary_debug_file)

            module.try_file(binary_path)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertEqual(module.wants_debug_file(), True)
            self.assertIsNone(module.debug_file_path)
            self.assertIsNone(module.supplementary_debug_file_kind)
            self.assertIsNone(module.supplementary_debug_file_path)
            self.assertEqual(
                module.wanted_supplementary_debug_file(),
                (
                    SupplementaryFileKind.GNU_DEBUGALTLINK,
                    str(binary_path),
                    str(alt_path),
                    alt_build_id,
                ),
            )

            with self.assertRaises(ValueError):
                module.debug_file_status = ModuleFileStatus.HAVE
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            module.debug_file_status = ModuleFileStatus.WANT_SUPPLEMENTARY
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )

            module.try_file(alt_path)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_path, str(binary_path))
            self.assertEqual(
                module.supplementary_debug_file_kind,
                SupplementaryFileKind.GNU_DEBUGALTLINK,
            )
            self.assertEqual(module.supplementary_debug_file_path, str(alt_path))
            self.assertRaises(ValueError, module.wanted_supplementary_debug_file)

    def test_gnu_debugaltlink_build_id_mismatch(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            alt_path = debug_dir / "alt.debug"
            alt_path.write_bytes(create_dwarf_file((), build_id=alt_build_id[::-1]))

            binary_path = bin_dir / "binary"
            binary_path.write_bytes(
                create_dwarf_file(
                    (),
                    sections=(ALLOCATED_SECTION,),
                    build_id=build_id,
                    gnu_debugaltlink=(alt_path, alt_build_id),
                )
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            module.build_id = build_id

            self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
            self.assertRaises(ValueError, module.wanted_supplementary_debug_file)

            module.try_file(binary_path)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertIsNone(module.debug_file_path)
            self.assertIsNone(module.supplementary_debug_file_kind)
            self.assertIsNone(module.supplementary_debug_file_path)
            self.assertEqual(
                module.wanted_supplementary_debug_file(),
                (
                    SupplementaryFileKind.GNU_DEBUGALTLINK,
                    str(binary_path),
                    str(alt_path),
                    alt_build_id,
                ),
            )

            module.try_file(alt_path)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertIsNone(module.debug_file_path)
            self.assertIsNone(module.supplementary_debug_file_kind)
            self.assertIsNone(module.supplementary_debug_file_path)
            self.assertEqual(
                module.wanted_supplementary_debug_file(),
                (
                    SupplementaryFileKind.GNU_DEBUGALTLINK,
                    str(binary_path),
                    str(alt_path),
                    alt_build_id,
                ),
            )

    def test_gnu_debugaltlink_then_both(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            alt_path = debug_dir / "alt.debug"
            alt_path.write_bytes(create_dwarf_file((), build_id=alt_build_id))

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            module.build_id = build_id
            with NamedTemporaryElfFile(
                build_id=build_id,
                gnu_debugaltlink=(alt_path, alt_build_id),
            ) as f1:
                module.try_file(f1.name)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )

            with NamedTemporaryElfFile(build_id=build_id) as f2:
                module.try_file(f2.name)

            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.loaded_file_path, f1.name)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_path, f2.name)

    def test_gnu_debugaltlink_cancel(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            alt_path = debug_dir / "alt.debug"
            alt_path.write_bytes(create_dwarf_file((), build_id=alt_build_id))

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            module.build_id = build_id
            with NamedTemporaryElfFile(
                build_id=build_id,
                gnu_debugaltlink=(alt_path, alt_build_id),
            ) as f:
                module.try_file(f.name)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )

            module.debug_file_status = ModuleFileStatus.WANT
            self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
            self.assertEqual(module.wants_debug_file(), True)
            self.assertRaises(ValueError, module.wanted_supplementary_debug_file)

    def test_extra_module_no_address_range(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        with NamedTemporaryElfFile() as f:
            module.try_file(f.name)
        self.assertIsNone(module.address_range)
        self.assertEqual(module.loaded_file_bias, 0)
        self.assertEqual(module.debug_file_bias, 0)

    def test_extra_module_address_range(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.address_range = (0x40000000, 0x40001000)
        with NamedTemporaryElfFile() as f:
            module.try_file(f.name)
        self.assertEqual(module.address_range, (0x40000000, 0x40001000))
        self.assertEqual(module.loaded_file_bias, 0x30000000)
        self.assertEqual(module.debug_file_bias, 0x30000000)

    def test_extra_module_empty_address_range(self):
        module = self.prog.extra_module("/foo/bar", create=True)
        module.address_range = (0, 0)
        with NamedTemporaryElfFile() as f:
            module.try_file(f.name)
        self.assertEqual(module.address_range, (0, 0))
        self.assertEqual(module.loaded_file_bias, 0)
        self.assertEqual(module.debug_file_bias, 0)


class TestLinuxUserspaceCoreDump(TestCase):
    def setUp(self):
        self.prog = Program()
        self.prog.debug_info_options.directories = ()
        self.prog.debug_info_options.debug_link_directories = ()
        self.prog.set_enabled_debug_info_finders(["standard"])

    def test_loaded_modules(self):
        self.prog.set_core_dump(get_resource("crashme.core"))

        loaded_modules = []
        for module, new in self.prog.loaded_modules():
            self.assertEqual(new, True)
            loaded_modules.append(module)
        found_modules = []

        with self.subTest(module="main"):
            module = self.prog.main_module()
            found_modules.append(module)
            self.assertEqual(module.name, "/home/osandov/crashme")
            self.assertEqual(module.address_range, (0x400000, 0x404010))
            self.assertEqual(
                module.build_id.hex(), "99a6524c4df01fbff9b43a6ead3d8e8e6201568b"
            )

        with self.subTest(module="crashme"):
            module = self.prog.shared_library_module(
                "/home/osandov/crashme.so", 0x7F6112CACE08
            )
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7F6112CA9000, 0x7F6112CAD010))
            self.assertEqual(
                module.build_id.hex(), "7bd58f10e741c3c8fbcf2031aa65f830f933d616"
            )

        with self.subTest(module="libc"):
            module = self.prog.shared_library_module("/lib64/libc.so.6", 0x7F6112C94960)
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7F6112AAE000, 0x7F6112C9EB70))
            self.assertEqual(
                module.build_id.hex(), "77c77fee058b19c6f001cf2cb0371ce3b8341211"
            )

        with self.subTest(module="ld-linux"):
            module = self.prog.shared_library_module(
                "/lib64/ld-linux-x86-64.so.2", 0x7F6112CEAE68
            )
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7F6112CB6000, 0x7F6112CEC2D8))
            self.assertEqual(
                module.build_id.hex(), "91dcd0244204201b616bbf59427771b3751736ce"
            )

        with self.subTest(module="vdso"):
            module = self.prog.vdso_module("linux-vdso.so.1", 0x7F6112CB4438)
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7F6112CB4000, 0x7F6112CB590F))
            self.assertEqual(
                module.build_id.hex(), "fdc3e4d463911345fbc6d9cc432e5ebc276e8e03"
            )

        self.assertCountEqual(loaded_modules, found_modules)

        loaded_modules = []
        for module, new in self.prog.loaded_modules():
            self.assertEqual(new, False)
            loaded_modules.append(module)
        self.assertCountEqual(loaded_modules, found_modules)

    def _try_vdso_in_core(self, module):
        module.debug_file_status = ModuleFileStatus.DONT_WANT
        self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)

    def test_bias(self):
        self.prog.set_core_dump(get_resource("crashme.core"))
        self.prog.create_loaded_modules()

        with self.subTest(module="main"):
            module = self.prog.main_module()
            module.try_file(get_resource("crashme"))
            self.assertEqual(module.loaded_file_bias, 0)
            self.assertEqual(module.debug_file_bias, 0)

        with self.subTest(module="crashme"):
            module = self.prog.shared_library_module(
                "/home/osandov/crashme.so", 0x7F6112CACE08
            )
            module.try_file(get_resource("crashme.so"))
            self.assertEqual(module.loaded_file_bias, 0x7F6112CA9000)
            self.assertEqual(module.debug_file_bias, 0x7F6112CA9000)

        with self.subTest(module="vdso"):
            module = self.prog.vdso_module("linux-vdso.so.1", 0x7F6112CB4438)
            self._try_vdso_in_core(module)
            self.assertEqual(module.loaded_file_bias, 0x7F6112CB4000)
            self.assertIsNone(module.debug_file_bias)

    def test_loaded_modules_pie(self):
        self.prog.set_core_dump(get_resource("crashme_pie.core"))

        loaded_modules = []
        for module, new in self.prog.loaded_modules():
            self.assertEqual(new, True)
            loaded_modules.append(module)
        found_modules = []

        with self.subTest(module="main"):
            module = self.prog.main_module()
            found_modules.append(module)
            self.assertEqual(module.name, "/home/osandov/crashme_pie")
            self.assertEqual(module.address_range, (0x557ED343D000, 0x557ED3441018))
            self.assertEqual(
                module.build_id.hex(), "eb4ad7aaded3815ab133a6d7784a2c95a4e52998"
            )

        with self.subTest(module="crashme"):
            module = self.prog.shared_library_module(
                "/home/osandov/crashme.so", 0x7FAB2C38DE08
            )
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7FAB2C38A000, 0x7FAB2C38E010))
            self.assertEqual(
                module.build_id.hex(), "7bd58f10e741c3c8fbcf2031aa65f830f933d616"
            )

        with self.subTest(module="libc"):
            module = self.prog.shared_library_module("/lib64/libc.so.6", 0x7FAB2C375960)
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7FAB2C18F000, 0x7FAB2C37FB70))
            self.assertEqual(
                module.build_id.hex(), "77c77fee058b19c6f001cf2cb0371ce3b8341211"
            )

        with self.subTest(module="ld-linux"):
            module = self.prog.shared_library_module(
                "/lib64/ld-linux-x86-64.so.2", 0x7FAB2C3CBE68
            )
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7FAB2C397000, 0x7FAB2C3CD2D8))
            self.assertEqual(
                module.build_id.hex(), "91dcd0244204201b616bbf59427771b3751736ce"
            )

        with self.subTest(module="vdso"):
            module = self.prog.vdso_module("linux-vdso.so.1", 0x7FAB2C395438)
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7FAB2C395000, 0x7FAB2C39690F))
            self.assertEqual(
                module.build_id.hex(), "fdc3e4d463911345fbc6d9cc432e5ebc276e8e03"
            )

        self.assertCountEqual(loaded_modules, found_modules)

        loaded_modules = []
        for module, new in self.prog.loaded_modules():
            self.assertEqual(new, False)
            loaded_modules.append(module)
        self.assertCountEqual(loaded_modules, found_modules)

    def test_bias_pie(self):
        self.prog.set_core_dump(get_resource("crashme_pie.core"))
        self.prog.create_loaded_modules()

        with self.subTest(module="main"):
            module = self.prog.main_module()
            module.try_file(get_resource("crashme_pie"))
            self.assertEqual(module.loaded_file_bias, 0x557ED343D000)
            self.assertEqual(module.debug_file_bias, 0x557ED343D000)

        with self.subTest(module="crashme"):
            module = self.prog.shared_library_module(
                "/home/osandov/crashme.so", 0x7FAB2C38DE08
            )
            module.try_file(get_resource("crashme.so"))
            self.assertEqual(module.loaded_file_bias, 0x7FAB2C38A000)
            self.assertEqual(module.debug_file_bias, 0x7FAB2C38A000)

        with self.subTest(module="vdso"):
            module = self.prog.vdso_module("linux-vdso.so.1", 0x7FAB2C395438)
            self._try_vdso_in_core(module)
            self.assertEqual(module.loaded_file_bias, 0x7FAB2C395000)
            self.assertIsNone(module.debug_file_bias)

    def test_loaded_modules_static(self):
        self.prog.set_core_dump(get_resource("crashme_static.core"))

        loaded_modules = []
        for module, new in self.prog.loaded_modules():
            self.assertEqual(new, True)
            loaded_modules.append(module)
        found_modules = []

        with self.subTest(module="main"):
            module = self.prog.main_module()
            found_modules.append(module)
            self.assertEqual(module.name, "/home/osandov/crashme_static")
            self.assertEqual(module.address_range, (0x400000, 0x4042B8))
            self.assertEqual(
                module.build_id.hex(), "a0b6befad9f0883c52c475ba3cee9c549cd082cf"
            )

        with self.subTest(module="vdso"):
            module = self.prog.vdso_module("linux-vdso.so.1", 0x7FBC73A66438)
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7FBC73A66000, 0x7FBC73A6790F))
            self.assertEqual(
                module.build_id.hex(), "fdc3e4d463911345fbc6d9cc432e5ebc276e8e03"
            )

        self.assertCountEqual(loaded_modules, found_modules)

        loaded_modules = []
        for module, new in self.prog.loaded_modules():
            self.assertEqual(new, False)
            loaded_modules.append(module)
        self.assertCountEqual(loaded_modules, found_modules)

    def test_bias_static(self):
        self.prog.set_core_dump(get_resource("crashme_static.core"))
        self.prog.create_loaded_modules()

        with self.subTest(module="main"):
            module = self.prog.main_module()
            module.try_file(get_resource("crashme_static"))
            self.assertEqual(module.loaded_file_bias, 0x0)
            self.assertEqual(module.debug_file_bias, 0x0)

        with self.subTest(module="vdso"):
            module = self.prog.vdso_module("linux-vdso.so.1", 0x7FBC73A66438)
            self._try_vdso_in_core(module)
            self.assertEqual(module.loaded_file_bias, 0x7FBC73A66000)
            self.assertIsNone(module.debug_file_bias)

    def test_loaded_modules_static_pie(self):
        self.prog.set_core_dump(get_resource("crashme_static_pie.core"))

        loaded_modules = []
        for module, new in self.prog.loaded_modules():
            self.assertEqual(new, True)
            loaded_modules.append(module)
        found_modules = []

        with self.subTest(module="main"):
            module = self.prog.main_module()
            found_modules.append(module)
            self.assertEqual(module.name, "/home/osandov/crashme_static_pie")
            self.assertEqual(module.address_range, (0x7FD981DC9000, 0x7FD981DCD278))
            self.assertEqual(
                module.build_id.hex(), "3e0bc47f80d7e64724e11fc021a251ed0d35bc2c"
            )

        with self.subTest(module="vdso"):
            module = self.prog.vdso_module("linux-vdso.so.1", 0x7FD981DC7438)
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7FD981DC7000, 0x7FD981DC890F))
            self.assertEqual(
                module.build_id.hex(), "fdc3e4d463911345fbc6d9cc432e5ebc276e8e03"
            )

        self.assertCountEqual(loaded_modules, found_modules)

        loaded_modules = []
        for module, new in self.prog.loaded_modules():
            self.assertEqual(new, False)
            loaded_modules.append(module)
        self.assertCountEqual(loaded_modules, found_modules)

    def test_bias_static_pie(self):
        self.prog.set_core_dump(get_resource("crashme_static_pie.core"))
        self.prog.create_loaded_modules()

        with self.subTest(module="main"):
            module = self.prog.main_module()
            module.try_file(get_resource("crashme_static_pie"))
            self.assertEqual(module.loaded_file_bias, 0x7FD981DC9000)
            self.assertEqual(module.debug_file_bias, 0x7FD981DC9000)

        with self.subTest(module="vdso"):
            module = self.prog.vdso_module("linux-vdso.so.1", 0x7FD981DC7438)
            self._try_vdso_in_core(module)
            self.assertEqual(module.loaded_file_bias, 0x7FD981DC7000)
            self.assertIsNone(module.debug_file_bias)

    def test_loaded_modules_pie_no_headers(self):
        self.prog.set_core_dump(get_resource("crashme_pie_no_headers.core"))

        loaded_modules = []
        for module, new in self.prog.loaded_modules():
            self.assertEqual(new, True)
            loaded_modules.append(module)
        found_modules = []

        # Without ELF headers saved in the core dump, and without the main ELF
        # file, only the main module (with limited information) and vDSO can be
        # found.
        with self.subTest(module="main"):
            module = self.prog.main_module()
            found_modules.append(module)
            self.assertEqual(module.name, "/home/osandov/crashme_pie")
            self.assertIsNone(module.address_range)
            self.assertIsNone(module.build_id)

        with self.subTest(module="vdso"):
            module = self.prog.vdso_module("linux-vdso.so.1", 0x7F299F607438)
            found_modules.append(module)
            self.assertEqual(module.address_range, (0x7F299F607000, 0x7F299F60890F))
            self.assertEqual(
                module.build_id.hex(), "fdc3e4d463911345fbc6d9cc432e5ebc276e8e03"
            )

        self.assertCountEqual(loaded_modules, found_modules)

        loaded_modules = []
        for module, new in self.prog.loaded_modules():
            self.assertEqual(new, False)
            loaded_modules.append(module)
        self.assertCountEqual(loaded_modules, found_modules)

        # If we can read the file headers (specifically, the program header
        # table and the interpreter path), then we should be able to get all of
        # the modules (with limited information).
        exe_file = self.enterContext(open(get_resource("crashme_pie"), "rb"))

        def read_headers(address, count, offset, physical):
            exe_file.seek(offset)
            return exe_file.read(count)

        self.prog.add_memory_segment(0x5623363D6000, 4096, read_headers, False)

        old_loaded_modules = []
        new_loaded_modules = []
        for module, new in self.prog.loaded_modules():
            (new_loaded_modules if new else old_loaded_modules).append(module)
        new_found_modules = []

        with self.subTest(module="main2"):
            module = self.prog.main_module()
            self.assertIsNone(module.address_range)
            self.assertIsNone(module.build_id)

        with self.subTest(module="crashme"):
            module = self.prog.shared_library_module(
                "/home/osandov/crashme.so", 0x7F299F5FFE08
            )
            new_found_modules.append(module)
            self.assertIsNone(module.address_range)
            self.assertIsNone(module.build_id)

        with self.subTest(module="libc"):
            module = self.prog.shared_library_module("/lib64/libc.so.6", 0x7F299F5E7960)
            new_found_modules.append(module)
            self.assertIsNone(module.address_range)
            self.assertIsNone(module.build_id)

        with self.subTest(module="ld-linux"):
            module = self.prog.shared_library_module(
                "/lib64/ld-linux-x86-64.so.2", 0x7F299F63DE68
            )
            new_found_modules.append(module)
            self.assertIsNone(module.address_range)
            self.assertIsNone(module.build_id)

        self.assertCountEqual(old_loaded_modules, loaded_modules)
        self.assertCountEqual(new_loaded_modules, new_found_modules)


class TestLoadDebugInfo(TestCase):
    def setUp(self):
        self.prog = Program()
        self.prog.set_core_dump(get_resource("crashme.core"))
        self.prog.set_enabled_debug_info_finders([])
        self.finder = unittest.mock.Mock()
        self.prog.register_debug_info_finder("mock", self.finder, enable_index=0)

    def test_nothing(self):
        self.prog.load_debug_info(None, default=False, main=False)
        self.assertFalse(list(self.prog.modules()))
        self.finder.assert_not_called()

    def test_empty_list(self):
        self.prog.load_debug_info([], default=False, main=False)
        self.assertFalse(list(self.prog.modules()))
        self.finder.assert_not_called()

    def test_no_such_file(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            self.prog.load_debug_info([Path(tmp_dir) / "file"])
        self.assertFalse(list(self.prog.modules()))
        self.finder.assert_not_called()

    def test_not_elf(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"hello, world\n")
            f.flush()
            self.prog.load_debug_info([f.name])
        self.assertFalse(list(self.prog.modules()))
        self.finder.assert_not_called()

    def test_no_build_id(self):
        with NamedTemporaryElfFile() as f:
            self.prog.load_debug_info([f.name])
        self.assertFalse(list(self.prog.modules()))
        self.finder.assert_not_called()

    def test_only_main_path(self):
        crashme_path = get_resource("crashme")

        self.prog.load_debug_info([crashme_path])

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The provided path should be used for the main module.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_path),
        )
        # Finders shouldn't be called.
        self.finder.assert_not_called()

    def test_only_paths(self):
        crashme_path = get_resource("crashme")
        crashme_so_path = get_resource("crashme.so")

        self.prog.load_debug_info([crashme_path, crashme_so_path])

        modules = list(self.prog.modules())
        # All loaded modules should be created.
        self.assertEqual(len(modules), 5)
        # The provided files should be used for their respective modules.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_path),
        )
        crashme_so_module = next(
            module for module in modules if module.name == "/home/osandov/crashme.so"
        )
        self.assertEqual(
            crashme_so_module.loaded_file_path,
            str(crashme_so_path),
        )
        self.assertEqual(
            crashme_so_module.debug_file_path,
            str(crashme_so_path),
        )
        # The rest should not have a file.
        for module in modules:
            if module.name not in ("/home/osandov/crashme", "/home/osandov/crashme.so"):
                self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
                self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        # Finders shouldn't be called.
        self.finder.assert_not_called()

    def test_main_by_path(self):
        crashme_path = get_resource("crashme")

        self.prog.load_debug_info([crashme_path], main=True)

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The provided path should be used for the main module.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_path),
        )
        # Finders shouldn't be called.
        self.finder.assert_not_called()

    def test_main_by_finder(self):
        crashme_path = get_resource("crashme")

        def finder(modules):
            for module in modules:
                if module.name == "/home/osandov/crashme":
                    module.try_file(crashme_path)

        self.finder.side_effect = finder

        self.prog.load_debug_info(main=True)

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The finder should be called and set the file for the main module.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_path),
        )
        self.finder.assert_called_once_with([self.prog.main_module()])

    def test_default_by_paths(self):
        crashme_path = get_resource("crashme")
        crashme_so_path = get_resource("crashme.so")

        self.assertRaises(
            MissingDebugInfoError,
            self.prog.load_debug_info,
            [crashme_path, crashme_so_path],
            default=True,
        )

        # All loaded modules should be created.
        modules = list(self.prog.modules())
        self.assertEqual(len(modules), 5)
        # The provided files should be used for their respective modules.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_path),
        )
        crashme_so_module = next(
            module for module in modules if module.name == "/home/osandov/crashme.so"
        )
        self.assertEqual(
            crashme_so_module.loaded_file_path,
            str(crashme_so_path),
        )
        self.assertEqual(
            crashme_so_module.debug_file_path,
            str(crashme_so_path),
        )
        # The rest should not have a file.
        missing_modules = []
        for module in modules:
            if module.name not in ("/home/osandov/crashme", "/home/osandov/crashme.so"):
                self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
                self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
                missing_modules.append(module)
        self.assertEqual(len(missing_modules), 3)
        # The finder should be called for the rest.
        self.finder.assert_called_once()
        self.assertCountEqual(self.finder.call_args[0][0], missing_modules)

    def test_default_by_finder(self):
        crashme_path = get_resource("crashme")
        crashme_so_path = get_resource("crashme.so")

        def finder(modules):
            for module in modules:
                if module.name == "/home/osandov/crashme":
                    module.try_file(crashme_path)
                elif module.name == "/home/osandov/crashme.so":
                    module.try_file(crashme_so_path)

        self.finder.side_effect = finder

        self.assertRaises(
            MissingDebugInfoError, self.prog.load_debug_info, default=True
        )

        # All loaded modules should be created.
        modules = list(self.prog.modules())
        self.assertEqual(len(modules), 5)
        # The finder should be called and set the files for the matching
        # modules.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_path),
        )
        crashme_so_module = next(
            module for module in modules if module.name == "/home/osandov/crashme.so"
        )
        self.assertEqual(
            crashme_so_module.loaded_file_path,
            str(crashme_so_path),
        )
        self.assertEqual(
            crashme_so_module.debug_file_path,
            str(crashme_so_path),
        )
        # The rest should not have a file.
        for module in modules:
            if module.name not in ("/home/osandov/crashme", "/home/osandov/crashme.so"):
                self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
                self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        # The finder should be called for all loaded modules.
        self.finder.assert_called_once()
        self.assertCountEqual(self.finder.call_args[0][0], modules)

    def test_main_gnu_debugaltlink_by_path(self):
        crashme_dwz_path = get_resource("crashme.dwz")
        crashme_alt_path = get_resource("crashme.alt")

        self.prog.load_debug_info([crashme_dwz_path, crashme_alt_path], main=True)

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The provided paths should be used for the main module.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().supplementary_debug_file_path,
            str(crashme_alt_path),
        )
        # Finders shouldn't be called.
        self.finder.assert_not_called()

    def test_main_gnu_debugaltlink_by_finder(self):
        crashme_dwz_path = get_resource("crashme.dwz")
        crashme_alt_path = get_resource("crashme.alt")

        def finder(modules):
            for module in modules:
                if module.name == "/home/osandov/crashme":
                    module.try_file(crashme_dwz_path)
                    module.try_file(crashme_alt_path)

        self.finder.side_effect = finder

        self.prog.load_debug_info(main=True)

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The finder should be called and set the files for the main module.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().supplementary_debug_file_path,
            str(crashme_alt_path),
        )
        self.finder.assert_called_once_with([self.prog.main_module()])

    def test_main_by_path_gnu_debugaltlink_not_found(self):
        crashme_dwz_path = get_resource("crashme.dwz")

        def finder(modules):
            for module in modules:
                if module.name == "/home/osandov/crashme":
                    self.assertEqual(
                        module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
                    )

        self.finder.side_effect = finder

        self.assertRaises(
            MissingDebugInfoError,
            self.prog.load_debug_info,
            [crashme_dwz_path],
            main=True,
        )

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The provided path should be used for the loaded file.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_dwz_path),
        )
        # The finder should be called and fail to find the supplementary file
        # for the main module.
        self.assertEqual(
            self.prog.main_module().debug_file_status,
            ModuleFileStatus.WANT_SUPPLEMENTARY,
        )
        self.assertEqual(
            self.prog.main_module().wanted_supplementary_debug_file()[:3],
            (
                SupplementaryFileKind.GNU_DEBUGALTLINK,
                str(crashme_dwz_path),
                "crashme.alt",
            ),
        )
        self.finder.assert_called_once_with([self.prog.main_module()])

    def test_main_by_finder_gnu_debugaltlink_not_found(self):
        crashme_dwz_path = get_resource("crashme.dwz")

        def finder(modules):
            for module in modules:
                if module.name == "/home/osandov/crashme":
                    self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
                    module.try_file(crashme_dwz_path)
                    self.assertEqual(
                        module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
                    )

        self.finder.side_effect = finder

        self.assertRaises(MissingDebugInfoError, self.prog.load_debug_info, main=True)

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The finder should be called and set the loaded file for the main
        # module but fail to find the supplementary file.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_status,
            ModuleFileStatus.WANT_SUPPLEMENTARY,
        )
        self.assertEqual(
            self.prog.main_module().wanted_supplementary_debug_file()[:3],
            (
                SupplementaryFileKind.GNU_DEBUGALTLINK,
                str(crashme_dwz_path),
                "crashme.alt",
            ),
        )
        self.finder.assert_called_once_with([self.prog.main_module()])

    def test_main_by_path_gnu_debugaltlink_by_finder(self):
        crashme_dwz_path = get_resource("crashme.dwz")
        crashme_alt_path = get_resource("crashme.alt")

        def finder(modules):
            for module in modules:
                if (
                    module.name == "/home/osandov/crashme"
                    and module.debug_file_status == ModuleFileStatus.WANT_SUPPLEMENTARY
                ):
                    module.try_file(crashme_alt_path)

        self.finder.side_effect = finder

        self.prog.load_debug_info([crashme_dwz_path], main=True)

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The provided path should be used for the main module.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_dwz_path),
        )
        # The finder should be called and set the supplementary file for the
        # main module.
        self.assertEqual(
            self.prog.main_module().supplementary_debug_file_path,
            str(crashme_alt_path),
        )
        self.finder.assert_called_once_with([self.prog.main_module()])

    def test_main_by_finder_gnu_debugaltlink_by_path(self):
        crashme_dwz_path = get_resource("crashme.dwz")
        crashme_alt_path = get_resource("crashme.alt")

        def finder(modules):
            for module in modules:
                if module.name == "/home/osandov/crashme":
                    module.try_file(crashme_dwz_path)

        self.finder.side_effect = finder

        self.prog.load_debug_info([crashme_alt_path], main=True)

        # The provided path should be used for the supplementary file for the
        # main module.
        self.assertEqual(
            self.prog.main_module().supplementary_debug_file_path,
            str(crashme_alt_path),
        )
        # The finder should be called and set the file for the main module.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_dwz_path),
        )
        self.finder.assert_called_once_with([self.prog.main_module()])

    def test_main_wants_gnu_debugaltlink_by_path(self):
        crashme_dwz_path = get_resource("crashme.dwz")
        crashme_alt_path = get_resource("crashme.alt")

        for module, _ in self.prog.loaded_modules():
            if isinstance(module, MainModule):
                module.try_file(crashme_dwz_path)
                break
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_status,
            ModuleFileStatus.WANT_SUPPLEMENTARY,
        )

        self.prog.load_debug_info([crashme_alt_path], main=True)

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The provided path should be used for the supplementary file.
        self.assertEqual(
            self.prog.main_module().supplementary_debug_file_path,
            str(crashme_alt_path),
        )
        # Finders shouldn't be called.
        self.finder.assert_not_called()

    def test_main_wants_gnu_debugaltlink_by_finder(self):
        crashme_dwz_path = get_resource("crashme.dwz")
        crashme_alt_path = get_resource("crashme.alt")

        for module, _ in self.prog.loaded_modules():
            if isinstance(module, MainModule):
                module.try_file(crashme_dwz_path)
                break
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_status,
            ModuleFileStatus.WANT_SUPPLEMENTARY,
        )

        def finder(modules):
            for module in modules:
                if (
                    module.name == "/home/osandov/crashme"
                    and module.debug_file_status == ModuleFileStatus.WANT_SUPPLEMENTARY
                ):
                    module.try_file(crashme_alt_path)

        self.finder.side_effect = finder

        self.prog.load_debug_info(main=True)

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The finder should be called and set the supplementary file for the
        # main module.
        self.assertEqual(
            self.prog.main_module().supplementary_debug_file_path,
            str(crashme_alt_path),
        )
        self.finder.assert_called_once_with([self.prog.main_module()])

    def test_main_wants_gnu_debugaltlink_not_found(self):
        crashme_dwz_path = get_resource("crashme.dwz")

        for module, _ in self.prog.loaded_modules():
            if isinstance(module, MainModule):
                module.try_file(crashme_dwz_path)
                break
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_status,
            ModuleFileStatus.WANT_SUPPLEMENTARY,
        )

        self.assertRaises(MissingDebugInfoError, self.prog.load_debug_info, main=True)

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The finder should be called and fail to find the supplementary file
        # for the main module, but the supplementary file should still be
        # wanted.
        self.assertEqual(
            self.prog.main_module().debug_file_status,
            ModuleFileStatus.WANT_SUPPLEMENTARY,
        )
        self.finder.assert_called_once_with([self.prog.main_module()])

    def test_default_gnu_debugaltlink_by_paths(self):
        crashme_dwz_path = get_resource("crashme.dwz")
        crashme_so_dwz_path = get_resource("crashme.so.dwz")
        crashme_alt_path = get_resource("crashme.alt")

        self.assertRaises(
            MissingDebugInfoError,
            self.prog.load_debug_info,
            [crashme_dwz_path, crashme_so_dwz_path, crashme_alt_path],
            default=True,
        )

        # All loaded modules should be created.
        modules = list(self.prog.modules())
        self.assertEqual(len(modules), 5)
        # The provided files should be used for their respective modules.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_dwz_path),
        )
        self.assertEqual(
            self.prog.main_module().supplementary_debug_file_path,
            str(crashme_alt_path),
        )
        crashme_so_module = next(
            module for module in modules if module.name == "/home/osandov/crashme.so"
        )
        self.assertEqual(
            crashme_so_module.loaded_file_path,
            str(crashme_so_dwz_path),
        )
        self.assertEqual(
            crashme_so_module.debug_file_path,
            str(crashme_so_dwz_path),
        )
        self.assertEqual(
            crashme_so_module.supplementary_debug_file_path,
            str(crashme_alt_path),
        )
        # The rest should not have a file.
        missing_modules = []
        for module in modules:
            if module.name not in ("/home/osandov/crashme", "/home/osandov/crashme.so"):
                self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
                self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
                missing_modules.append(module)
        self.assertEqual(len(missing_modules), 3)
        # The finder should be called for the rest.
        self.finder.assert_called_once()
        self.assertCountEqual(self.finder.call_args[0][0], missing_modules)

    def test_dont_want(self):
        for module, _ in self.prog.loaded_modules():
            if isinstance(module, MainModule):
                module.loaded_file_status = ModuleFileStatus.DONT_WANT
                module.debug_file_status = ModuleFileStatus.DONT_WANT
                break
        # DONT_WANT should be reset to WANT.
        self.assertRaises(MissingDebugInfoError, self.prog.load_debug_info, main=True)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.finder.assert_called_once_with([self.prog.main_module()])

    def test_dont_need(self):
        for module, _ in self.prog.loaded_modules():
            if isinstance(module, MainModule):
                module.loaded_file_status = ModuleFileStatus.DONT_NEED
                module.debug_file_status = ModuleFileStatus.DONT_NEED
                break
        # DONT_NEED should be preserved.
        self.prog.load_debug_info(main=True)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.DONT_NEED)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.DONT_NEED)
        self.finder.assert_not_called()

    def test_unmatched(self):
        self.prog.load_debug_info([get_resource("crashme_static")])
        modules = list(self.prog.modules())
        # All loaded modules should be created.
        self.assertEqual(len(modules), 5)
        # None of them should have files.
        for module in modules:
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.finder.assert_not_called()


class TestLoadDebugInfoCoreNoHeaders(TestCase):
    def setUp(self):
        self.prog = Program()
        self.prog.set_core_dump(get_resource("crashme_pie_no_headers.core"))
        self.prog.set_enabled_debug_info_finders([])
        self.finder = unittest.mock.Mock()
        self.prog.register_debug_info_finder("mock", self.finder, enable_index=0)

    def test_main_by_finder(self):
        crashme_pie_path = get_resource("crashme_pie")

        def finder(modules):
            for module in modules:
                if module.name == "/home/osandov/crashme_pie":
                    module.try_file(crashme_pie_path)

        self.finder.side_effect = finder

        self.prog.load_debug_info(main=True)

        # The main module should be created.
        self.assertIn(self.prog.main_module(), list(self.prog.modules()))
        # The finder should be called and set the files, address range, and
        # build ID for the main module.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_pie_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_pie_path),
        )
        self.assertEqual(
            self.prog.main_module().address_range, (0x5623363D6000, 0x5623363DA018)
        )
        self.assertEqual(
            self.prog.main_module().build_id.hex(),
            "eb4ad7aaded3815ab133a6d7784a2c95a4e52998",
        )
        self.finder.assert_called_once_with([self.prog.main_module()])

    @unittest.expectedFailure  # Issue #291
    def test_default_by_finder(self):
        crashme_pie_path = get_resource("crashme_pie")
        crashme_so_path = get_resource("crashme.so")

        def finder(modules):
            for module in modules:
                if module.name == "/home/osandov/crashme_pie":
                    module.try_file(crashme_pie_path)
                elif module.name == "/home/osandov/crashme.so":
                    module.try_file(crashme_so_path)
                else:
                    module.loaded_file_status = ModuleFileStatus.DONT_NEED
                    module.debug_file_status = ModuleFileStatus.DONT_NEED

        self.finder.side_effect = finder

        self.prog.load_debug_info(default=True)

        # All loaded modules should be created (except ld-linux.so; see
        # tests.test_module.TestLinuxUserspaceCoreDump.test_loaded_modules_pie_no_headers).
        self.assertCountEqual(
            list(self.prog.modules()),
            [
                self.prog.main_module(),
                self.prog.vdso_module("linux-vdso.so.1", 0x7F299F607438),
                self.prog.shared_library_module(
                    "/home/osandov/crashme.so", 0x7F299F5FFE08
                ),
                self.prog.shared_library_module("/lib64/libc.so.6", 0x7F299F5E7960),
                self.prog.shared_library_module(
                    "/lib64/ld-linux-x86-64.so.2", 0x7F299F63DE68
                ),
            ],
        )
        # The finder should be called and set the files, address range, and
        # build ID for the main and crashme.so modules.
        self.assertEqual(
            self.prog.main_module().loaded_file_path,
            str(crashme_pie_path),
        )
        self.assertEqual(
            self.prog.main_module().debug_file_path,
            str(crashme_pie_path),
        )
        self.assertEqual(
            self.prog.main_module().address_range, (0x5623363D6000, 0x5623363DA018)
        )
        self.assertEqual(
            self.prog.main_module().build_id.hex(),
            "eb4ad7aaded3815ab133a6d7784a2c95a4e52998",
        )
        self.assertEqual(
            self.prog.shared_library_module(
                "/home/osandov/crashme.so", 0x7F299F5FFE08
            ).loaded_file_path,
            str(crashme_so_path),
        )
        self.assertEqual(
            self.prog.shared_library_module(
                "/home/osandov/crashme.so", 0x7F299F5FFE08
            ).debug_file_path,
            str(crashme_so_path),
        )
        self.assertEqual(
            self.prog.shared_library_module(
                "/home/osandov/crashme.so", 0x7F299F5FFE08
            ).address_range,
            (0x7F299F5FC000, 0x7F299F600010),
        )
        self.assertEqual(
            self.prog.shared_library_module(
                "/home/osandov/crashme.so", 0x7F299F5FFE08
            ).build_id.hex(),
            "7bd58f10e741c3c8fbcf2031aa65f830f933d616",
        )
        self.finder.assert_called()


class TestLoadModuleDebugInfo(TestCase):
    def setUp(self):
        self.prog = Program()
        self.prog.set_enabled_debug_info_finders([])
        self.finder = unittest.mock.Mock()
        self.prog.register_debug_info_finder("mock", self.finder, enable_index=0)

    def test_empty(self):
        self.prog.load_module_debug_info()
        self.finder.assert_not_called()

    def test_multiple(self):
        self.prog.load_module_debug_info(
            self.prog.extra_module("/foo/bar", create=True),
            self.prog.extra_module("/foo/baz", create=True),
        )
        self.finder.assert_called_once()
        self.assertCountEqual(
            self.finder.call_args[0][0],
            [
                self.prog.extra_module("/foo/bar"),
                self.prog.extra_module("/foo/baz"),
            ],
        )

    def test_wrong_program(self):
        self.assertRaisesRegex(
            ValueError,
            "module from wrong program",
            self.prog.load_module_debug_info,
            self.prog.extra_module("/foo/bar", create=True),
            Program().extra_module("/foo/baz", create=True),
        )

    def test_type_error(self):
        self.assertRaises(
            TypeError,
            self.prog.load_module_debug_info,
            self.prog.extra_module("/foo/bar", create=True),
            None,
        )


class TestStandardDebugInfoFinder(TestCase):
    def setUp(self):
        self.prog = Program()
        self.prog.debug_info_options.directories = ()
        self.prog.debug_info_options.debug_link_directories = ()
        self.prog.set_enabled_debug_info_finders(["standard"])

    def test_by_module_name(self):
        with NamedTemporaryElfFile() as f:
            module = self.prog.extra_module(f.name, create=True)
            self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f.name)
        self.assertEqual(module.debug_file_path, f.name)

    def test_by_module_name_with_build_id(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        with NamedTemporaryElfFile(build_id=build_id) as f:
            module = self.prog.extra_module(f.name, create=True)
            module.build_id = build_id
            self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f.name)
        self.assertEqual(module.debug_file_path, f.name)

    def test_by_module_name_missing_build_id(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        with NamedTemporaryElfFile() as f:
            module = self.prog.extra_module(f.name, create=True)
            module.build_id = build_id
            self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)

    def test_by_module_name_build_id_mismatch(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        with NamedTemporaryElfFile(build_id=build_id[::-1]) as f:
            module = self.prog.extra_module(f.name, create=True)
            module.build_id = build_id
            self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)

    def test_reuse_loaded_file(self):
        with NamedTemporaryElfFile() as f:
            module = self.prog.extra_module(f.name, create=True)
            module.debug_file_status = ModuleFileStatus.DONT_WANT
            self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.DONT_WANT)

        module.debug_file_status = ModuleFileStatus.WANT
        self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)

    def test_reuse_debug_file(self):
        with NamedTemporaryElfFile() as f:
            module = self.prog.extra_module(f.name, create=True)
            module.loaded_file_status = ModuleFileStatus.DONT_WANT
            self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.DONT_WANT)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)

        module.loaded_file_status = ModuleFileStatus.WANT
        self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.debug_file_path, f.name)

    def test_reuse_wanted_supplementary_debug_file(self):
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with NamedTemporaryElfFile(
            gnu_debugaltlink=("alt.debug", alt_build_id),
        ) as f:
            module = self.prog.extra_module(f.name, create=True)
            module.loaded_file_status = ModuleFileStatus.DONT_WANT
            self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.DONT_WANT)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY)

        module.loaded_file_status = ModuleFileStatus.WANT
        self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, f.name)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY)

    def test_vdso_in_core(self):
        self.prog.set_core_dump(get_resource("crashme.core"))
        for module, _ in self.prog.loaded_modules():
            if isinstance(module, VdsoModule):
                break
        else:
            self.fail("vDSO module not found")
        self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, "[vdso]")
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)

    def test_main_by_proc(self):
        self.prog.set_pid(os.getpid())
        for module, _ in self.prog.loaded_modules():
            if isinstance(module, MainModule):
                break
        else:
            self.fail("main module not found")
        self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)

    def test_vdso_by_proc(self):
        self.prog.set_pid(os.getpid())
        for module, _ in self.prog.loaded_modules():
            if isinstance(module, VdsoModule):
                break
        else:
            self.skipTest("vDSO module not found")
        self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
        self.assertEqual(module.loaded_file_path, "[vdso]")
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)

    def test_shared_library_by_proc(self):
        self.prog.set_pid(os.getpid())
        for module, _ in self.prog.loaded_modules():
            if isinstance(module, SharedLibraryModule):
                break
        else:
            self.skipTest("shared library module not found")
        self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)

    def test_by_build_id(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        for i, relative in enumerate((False, True)):
            with self.subTest(relative=relative):
                with tempfile.TemporaryDirectory(
                    prefix="bin-"
                ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
                    bin_dir = Path(bin_dir)
                    debug_dir = Path(debug_dir)

                    build_id_dir = debug_dir / ".build-id" / build_id.hex()[:2]
                    build_id_dir.mkdir(parents=True)
                    binary_path = build_id_dir / build_id.hex()[2:]
                    binary_path.write_bytes(
                        create_dwarf_file((), sections=(ALLOCATED_SECTION,))
                    )

                    module = self.prog.extra_module(bin_dir / "binary", i, create=True)
                    module.build_id = build_id

                    self.prog.debug_info_options.directories = (
                        os.path.relpath(debug_dir) if relative else str(debug_dir),
                    )
                    self.prog.load_module_debug_info(module)
                    self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
                    self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
                    self.assertEqual(module.loaded_file_path, str(binary_path))
                    self.assertEqual(module.debug_file_path, str(binary_path))

    def test_by_build_id_separate(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            build_id_dir = debug_dir / ".build-id" / build_id.hex()[:2]
            build_id_dir.mkdir(parents=True)
            loadable_path = build_id_dir / build_id.hex()[2:]
            loadable_path.write_bytes(
                create_elf_file(ET.EXEC, sections=(ALLOCATED_SECTION,))
            )
            debug_path = build_id_dir / (build_id.hex()[2:] + ".debug")
            debug_path.write_bytes(create_dwarf_file(()))

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            module.build_id = build_id

            self.prog.debug_info_options.directories = (str(debug_dir),)
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.loaded_file_path, str(loadable_path))
            self.assertEqual(module.debug_file_path, str(debug_path))

    def test_by_build_id_from_loaded(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            loadable_path = bin_dir / "binary"
            loadable_path.write_bytes(
                create_elf_file(
                    ET.EXEC, sections=(ALLOCATED_SECTION,), build_id=build_id
                )
            )
            build_id_dir = debug_dir / ".build-id" / build_id.hex()[:2]
            build_id_dir.mkdir(parents=True)
            debug_path = build_id_dir / (build_id.hex()[2:] + ".debug")
            debug_path.write_bytes(create_dwarf_file(()))

            module = self.prog.extra_module(bin_dir / "binary", create=True)

            self.prog.debug_info_options.directories = (str(debug_dir),)
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.loaded_file_path, str(loadable_path))
            self.assertEqual(module.debug_file_path, str(debug_path))

    def test_by_build_id_method(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            build_id_dir = debug_dir / ".build-id" / build_id.hex()[:2]
            build_id_dir.mkdir(parents=True)
            binary_path = build_id_dir / build_id.hex()[2:]
            binary_path.write_bytes(
                create_dwarf_file((), sections=(ALLOCATED_SECTION,))
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            module.build_id = build_id

            self.prog.find_standard_debug_info(
                [module],
                options=DebugInfoOptions(directories=(str(debug_dir),)),
            )
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.loaded_file_path, str(binary_path))
            self.assertEqual(module.debug_file_path, str(binary_path))

    def test_by_gnu_debuglink(self):
        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            debug_file_contents = create_dwarf_file(())
            crc = binascii.crc32(debug_file_contents)

            loadable_path = bin_dir / "binary"
            loadable_path.write_bytes(
                create_elf_file(
                    ET.EXEC,
                    sections=(ALLOCATED_SECTION,),
                    gnu_debuglink=("binary.debug", crc),
                )
            )

            self.prog.debug_info_options.directories = (str(debug_dir),)
            self.prog.debug_info_options.debug_link_directories = (
                "$ORIGIN",
                "$ORIGIN/.debug",
                "",
            )
            for i, debug_path in enumerate(
                (
                    bin_dir / "binary.debug",
                    bin_dir / ".debug" / "binary.debug",
                    debug_dir / bin_dir.relative_to("/") / "binary.debug",
                )
            ):
                with self.subTest(debug_path=debug_path):
                    try:
                        debug_path.parent.mkdir(parents=True, exist_ok=True)
                        debug_path.write_bytes(debug_file_contents)

                        module = self.prog.extra_module(
                            bin_dir / "binary", i, create=True
                        )

                        self.prog.load_module_debug_info(module)
                        self.assertEqual(
                            module.loaded_file_status, ModuleFileStatus.HAVE
                        )
                        self.assertEqual(
                            module.debug_file_status, ModuleFileStatus.HAVE
                        )
                        self.assertEqual(module.loaded_file_path, str(loadable_path))
                        self.assertEqual(module.debug_file_path, str(debug_path))
                    finally:
                        try:
                            debug_path.unlink()
                        except FileNotFoundError:
                            pass

    def test_by_gnu_debuglink_absolute(self):
        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            debug_file_contents = create_dwarf_file(())
            crc = binascii.crc32(debug_file_contents)
            debug_path = debug_dir / "binary.debug"

            loadable_path = bin_dir / "binary"
            loadable_path.write_bytes(
                create_elf_file(
                    ET.EXEC,
                    sections=(ALLOCATED_SECTION,),
                    gnu_debuglink=(debug_path, crc),
                )
            )

            debug_path.parent.mkdir(parents=True, exist_ok=True)
            debug_path.write_bytes(debug_file_contents)

            module = self.prog.extra_module(bin_dir / "binary", create=True)

            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.loaded_file_path, str(loadable_path))
            self.assertEqual(module.debug_file_path, str(debug_path))

    def test_by_gnu_debuglink_origin_with_braces(self):
        with tempfile.TemporaryDirectory(prefix="bin-") as bin_dir:
            bin_dir = Path(bin_dir)

            debug_file_contents = create_dwarf_file(())
            crc = binascii.crc32(debug_file_contents)
            debug_path = bin_dir / "binary.debug"

            loadable_path = bin_dir / "binary"
            loadable_path.write_bytes(
                create_elf_file(
                    ET.EXEC,
                    sections=(ALLOCATED_SECTION,),
                    gnu_debuglink=("binary.debug", crc),
                )
            )

            debug_path.parent.mkdir(parents=True, exist_ok=True)
            debug_path.write_bytes(debug_file_contents)

            module = self.prog.extra_module(bin_dir / "binary", create=True)

            self.prog.debug_info_options.debug_link_directories = ("${ORIGIN}",)
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.loaded_file_path, str(loadable_path))
            self.assertEqual(module.debug_file_path, str(debug_path))

    def test_by_gnu_debuglink_not_origin(self):
        # Test that strings other than $ORIGIN followed by a word boundary are
        # not replaced.
        for i, subdir in enumerate(("$ORIGINAL", "$foo", "$")):
            with self.subTest(subdir=subdir):
                with tempfile.TemporaryDirectory(prefix="bin-") as bin_dir:
                    bin_dir = Path(bin_dir)
                    debug_dir = bin_dir / subdir

                    debug_file_contents = create_dwarf_file(())
                    crc = binascii.crc32(debug_file_contents)
                    debug_path = debug_dir / "binary.debug"

                    loadable_path = bin_dir / "binary"
                    loadable_path.write_bytes(
                        create_elf_file(
                            ET.EXEC,
                            sections=(ALLOCATED_SECTION,),
                            gnu_debuglink=("binary.debug", crc),
                        )
                    )

                    debug_path.parent.mkdir(parents=True, exist_ok=True)
                    debug_path.write_bytes(debug_file_contents)

                    module = self.prog.extra_module(bin_dir / "binary", i, create=True)

                    self.prog.debug_info_options.debug_link_directories = (
                        str(debug_dir),
                    )
                    self.prog.load_module_debug_info(module)
                    self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
                    self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
                    self.assertEqual(module.loaded_file_path, str(loadable_path))
                    self.assertEqual(module.debug_file_path, str(debug_path))

    def test_by_gnu_debuglink_origin_multiple(self):
        # Pathological case combining the cases above.
        with tempfile.TemporaryDirectory(prefix="bin-") as bin_dir:
            bin_dir = Path(bin_dir)
            debug_dir = (
                bin_dir
                / "$ORIGINx"
                / (bin_dir.parent / (bin_dir.name + "x")).relative_to("/")
                / "$"
                / bin_dir.relative_to("/")
                / "$O"
            )

            debug_file_contents = create_dwarf_file(())
            crc = binascii.crc32(debug_file_contents)
            debug_path = debug_dir / "binary.debug"

            loadable_path = bin_dir / "binary"
            loadable_path.write_bytes(
                create_elf_file(
                    ET.EXEC,
                    sections=(ALLOCATED_SECTION,),
                    gnu_debuglink=("binary.debug", crc),
                )
            )

            debug_path.parent.mkdir(parents=True, exist_ok=True)
            debug_path.write_bytes(debug_file_contents)

            module = self.prog.extra_module(bin_dir / "binary", create=True)

            self.prog.debug_info_options.debug_link_directories = (
                str(bin_dir) + "/$ORIGINx${ORIGIN}x/$$ORIGIN/$O",
            )
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.loaded_file_path, str(loadable_path))
            self.assertEqual(module.debug_file_path, str(debug_path))

    def test_by_gnu_debuglink_crc_mismatch(self):
        with tempfile.TemporaryDirectory(prefix="bin-") as bin_dir:
            bin_dir = Path(bin_dir)

            debug_file_contents = create_dwarf_file(())
            crc = binascii.crc32(debug_file_contents)

            loadable_path = bin_dir / "binary"
            loadable_path.write_bytes(
                create_elf_file(
                    ET.EXEC,
                    sections=(ALLOCATED_SECTION,),
                    gnu_debuglink=("binary.debug", crc ^ 1),
                )
            )

            debug_path = bin_dir / "binary.debug"
            debug_path.write_bytes(debug_file_contents)

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            self.prog.debug_info_options.debug_link_directories = ("$ORIGIN",)
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)

    def test_invalid_gnu_debuglink(self):
        with tempfile.TemporaryDirectory(prefix="bin-") as bin_dir:
            bin_dir = Path(bin_dir)

            loadable_path = bin_dir / "binary"
            loadable_path.write_bytes(
                create_elf_file(
                    ET.EXEC,
                    sections=(
                        ALLOCATED_SECTION,
                        ElfSection(
                            name=".gnu_debuglink", sh_type=SHT.PROGBITS, data=b"foo"
                        ),
                    ),
                )
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
            self.assertEqual(module.loaded_file_path, str(loadable_path))

    def test_gnu_debugaltlink_absolute(self):
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            alt_path = debug_dir / "alt.debug"
            alt_path.write_bytes(create_dwarf_file((), build_id=alt_build_id))

            binary_path = bin_dir / "binary"
            binary_path.write_bytes(
                create_dwarf_file(
                    (),
                    sections=(ALLOCATED_SECTION,),
                    gnu_debugaltlink=(alt_path, alt_build_id),
                )
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.loaded_file_path, str(binary_path))
            self.assertEqual(module.debug_file_path, str(binary_path))
            self.assertEqual(module.supplementary_debug_file_path, str(alt_path))

    def test_gnu_debugaltlink_not_found(self):
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            binary_path = bin_dir / "binary"
            binary_path.write_bytes(
                create_dwarf_file(
                    (),
                    sections=(ALLOCATED_SECTION,),
                    gnu_debugaltlink=(debug_dir / "alt.debug", alt_build_id),
                )
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertEqual(
                module.wanted_supplementary_debug_file(),
                (
                    SupplementaryFileKind.GNU_DEBUGALTLINK,
                    str(binary_path),
                    str(debug_dir / "alt.debug"),
                    alt_build_id,
                ),
            )
            self.assertEqual(module.loaded_file_path, str(binary_path))

    def test_only_gnu_debugaltlink_absolute(self):
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            alt_path = debug_dir / "alt.debug"
            alt_path.write_bytes(create_dwarf_file((), build_id=alt_build_id))

            binary_path = bin_dir / "binary"
            binary_path.write_bytes(
                create_dwarf_file(
                    (),
                    sections=(ALLOCATED_SECTION,),
                    gnu_debugaltlink=(alt_path, alt_build_id),
                )
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            module.try_file(binary_path)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertEqual(module.loaded_file_path, str(binary_path))

            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_path, str(binary_path))
            self.assertEqual(module.supplementary_debug_file_path, str(alt_path))

    def test_only_gnu_debugaltlink_not_found(self):
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            binary_path = bin_dir / "binary"
            binary_path.write_bytes(
                create_dwarf_file(
                    (),
                    sections=(ALLOCATED_SECTION,),
                    gnu_debugaltlink=(debug_dir / "alt.debug", alt_build_id),
                )
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            module.try_file(binary_path)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertEqual(module.loaded_file_path, str(binary_path))

            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertEqual(
                module.wanted_supplementary_debug_file(),
                (
                    SupplementaryFileKind.GNU_DEBUGALTLINK,
                    str(binary_path),
                    str(debug_dir / "alt.debug"),
                    alt_build_id,
                ),
            )

    def test_gnu_debugaltlink_relative(self):
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            alt_path = debug_dir / "alt.debug"
            alt_path.write_bytes(create_dwarf_file((), build_id=alt_build_id))

            binary_path = bin_dir / "binary"
            binary_path.write_bytes(
                create_dwarf_file(
                    (),
                    sections=(ALLOCATED_SECTION,),
                    gnu_debugaltlink=(
                        Path(os.path.relpath(alt_path, bin_dir)),
                        alt_build_id,
                    ),
                )
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.loaded_file_path, str(binary_path))
            self.assertEqual(module.debug_file_path, str(binary_path))
            self.assertEqual(module.supplementary_debug_file_path, str(alt_path))

    def test_gnu_debugaltlink_debug_directories(self):
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            alt_path = debug_dir / ".dwz/alt.debug"
            alt_path.parent.mkdir()
            alt_path.write_bytes(create_dwarf_file((), build_id=alt_build_id))

            self.prog.debug_info_options.directories = (str(debug_dir),)
            for i, debugaltlink in enumerate(
                (
                    bin_dir / "debug/.dwz/alt.debug",
                    Path("debug/.dwz/alt.debug"),
                )
            ):
                with self.subTest(debugaltlink=debugaltlink):
                    binary_path = bin_dir / f"binary{i}"
                    binary_path.write_bytes(
                        create_dwarf_file(
                            (),
                            sections=(ALLOCATED_SECTION,),
                            gnu_debugaltlink=(debugaltlink, alt_build_id),
                        )
                    )

                    module = self.prog.extra_module(binary_path, create=True)
                    self.prog.load_module_debug_info(module)
                    self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
                    self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
                    self.assertEqual(module.loaded_file_path, str(binary_path))
                    self.assertEqual(module.debug_file_path, str(binary_path))
                    self.assertEqual(
                        module.supplementary_debug_file_path, str(alt_path)
                    )

    def test_gnu_debugaltlink_build_id_mismatch(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with tempfile.TemporaryDirectory(
            prefix="bin-"
        ) as bin_dir, tempfile.TemporaryDirectory(prefix="debug-") as debug_dir:
            bin_dir = Path(bin_dir)
            debug_dir = Path(debug_dir)

            alt_path = debug_dir / "alt.debug"
            alt_path.write_bytes(create_dwarf_file((), build_id=alt_build_id[::-1]))

            binary_path = bin_dir / "binary"
            binary_path.write_bytes(
                create_dwarf_file(
                    (),
                    sections=(ALLOCATED_SECTION,),
                    build_id=build_id,
                    gnu_debugaltlink=(alt_path, alt_build_id),
                )
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertEqual(
                module.wanted_supplementary_debug_file(),
                (
                    SupplementaryFileKind.GNU_DEBUGALTLINK,
                    str(binary_path),
                    str(alt_path),
                    alt_build_id,
                ),
            )
            self.assertEqual(module.loaded_file_path, str(binary_path))

    def test_invalid_gnu_debugaltlink(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        with tempfile.TemporaryDirectory(prefix="bin-") as bin_dir:
            bin_dir = Path(bin_dir)

            binary_path = bin_dir / "binary"
            binary_path.write_bytes(
                create_dwarf_file(
                    (),
                    sections=(
                        ALLOCATED_SECTION,
                        ElfSection(
                            name=".gnu_debugaltlink",
                            sh_type=SHT.PROGBITS,
                            data=b"foo",
                        ),
                    ),
                    build_id=build_id,
                )
            )

            module = self.prog.extra_module(bin_dir / "binary", create=True)
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
            self.assertEqual(module.loaded_file_path, str(binary_path))



# Copyright (c) 2026 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
import drgn.commands._crash  # noqa  # needed to register crash commands
from drgn.helpers.linux.cpumask import for_each_online_cpu
from tests.linux_kernel import skip_unless_have_stack_tracing
from tests.linux_kernel.vmcore import LinuxVMCoreCrashCommandTestCase


@skip_unless_have_stack_tracing
class TestBtVMCore(LinuxVMCoreCrashCommandTestCase):
    # When CPU 0 panics on s390x, the kernel switches to a different stack that
    # we don't know how to unwind through. This is a bug. For now,
    # vmtest.enter_kdump tries to avoid this so we can get some test coverage,
    # and we skip the tests otherwise.
    def _skip_if_cpu0_on_s390x(self):
        if NORMALIZED_MACHINE_NAME == "s390x" and self.prog["panic_cpu"].counter == 0:
            self.skipTest("drgn can't unwind s390x panic stack on CPU 0")

    def test_cpu(self):
        self._skip_if_cpu0_on_s390x()
        cmd = self.check_crash_command("bt -c 0")
        self.assertIn("CPU: 0", cmd.stdout)
        self.assertNotIn("(active)", cmd.stdout)

    def test_panic(self):
        self._skip_if_cpu0_on_s390x()
        tid = self.prog.crashed_thread().tid
        for cmd in ("bt", "bt -p"):
            with self.subTest(cmd=cmd):
                res = self.check_crash_command(cmd)
                self.assertIn("drgn_test_crash_func", res.stdout)
                self.assertIn(f"PID: {tid}", res.stdout)
                self.assertNotIn("(active)", res.stdout)

    def test_all(self):
        self._skip_if_cpu0_on_s390x()
        cmd = self.check_crash_command("bt -a")
        self.assertNotIn("(active)", cmd.stdout)
        for cpu in for_each_online_cpu(self.prog):
            self.assertIn(f"CPU: {cpu}", cmd.stdout)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import collections

from drgn import NULL, container_of
from drgn.helpers import ValidationError
from drgn.helpers.linux.rbtree import (
    RB_EMPTY_NODE,
    RB_EMPTY_ROOT,
    rb_find,
    rb_first,
    rb_last,
    rb_next,
    rb_parent,
    rb_prev,
    rbtree_inorder_for_each,
    rbtree_inorder_for_each_entry,
    rbtree_preorder_for_each,
    rbtree_preorder_for_each_entry,
    validate_rbtree,
    validate_rbtree_inorder_for_each_entry,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestRbtree(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.root = cls.prog["drgn_test_rb_root"].address_of_()
        cls.entries = cls.prog["drgn_test_rb_entries"]
        cls.num_entries = 4
        cls.empty_root = cls.prog["drgn_test_empty_rb_root"].address_of_()

    def node(self, n):
        return self.entries[n].node.address_of_()

    def entry(self, n):
        return self.entries[n].address_of_()

    def test_RB_EMPTY_ROOT(self):
        self.assertTrue(RB_EMPTY_ROOT(self.empty_root))
        self.assertFalse(RB_EMPTY_ROOT(self.root))

    def test_RB_EMPTY_NODE(self):
        self.assertTrue(
            RB_EMPTY_NODE(self.prog["drgn_test_empty_rb_node"].address_of_())
        )
        self.assertFalse(RB_EMPTY_NODE(self.node(0)))

    def test_rb_parent(self):
        if self.root.rb_node.rb_left:
            self.assertEqual(rb_parent(self.root.rb_node.rb_left), self.root.rb_node)
        if self.root.rb_node.rb_right:
            self.assertEqual(rb_parent(self.root.rb_node.rb_right), self.root.rb_node)

    def test_rb_first(self):
        self.assertEqual(rb_first(self.root), self.node(0))

    def test_rb_last(self):
        self.assertEqual(rb_last(self.root), self.node(self.num_entries - 1))

    def test_rb_next(self):
        for i in range(self.num_entries - 1):
            self.assertEqual(rb_next(self.node(i)), self.node(i + 1))
        self.assertEqual(
            rb_next(self.node(self.num_entries - 1)),
            NULL(self.prog, "struct rb_node *"),
        )

    def test_rb_prev(self):
        for i in range(1, self.num_entries):
            self.assertEqual(rb_prev(self.node(i)), self.node(i - 1))
        self.assertEqual(rb_prev(self.node(0)), NULL(self.prog, "struct rb_node *"))

    def test_rbtree_inorder_for_each(self):
        self.assertEqual(
            list(rbtree_inorder_for_each(self.root)),
            [self.node(i) for i in range(self.num_entries)],
        )

    def test_rbtree_inorder_for_each_entry(self):
        self.assertEqual(
            list(
                rbtree_inorder_for_each_entry(
                    "struct drgn_test_rb_entry", self.root, "node"
                )
            ),
            [self.entry(i) for i in range(self.num_entries)],
        )

    def test_rbtree_preorder_for_each(self):
        preorder = list(rbtree_preorder_for_each(self.root))
        self.assertCountEqual(preorder, [self.node(i) for i in range(self.num_entries)])
        self.assertEqual(preorder[0], self.root.rb_node)

    def test_rbtree_preorder_for_each_entry(self):
        preorder = list(
            rbtree_preorder_for_each_entry(
                "struct drgn_test_rb_entry", self.root, "node"
            )
        )
        self.assertCountEqual(
            preorder, [self.entry(i) for i in range(self.num_entries)]
        )
        self.assertEqual(
            preorder[0],
            container_of(self.root.rb_node, "struct drgn_test_rb_entry", "node"),
        )

    def test_rb_find(self):
        def cmp(key, obj):
            value = obj.value.value_()
            return key - value

        for i in range(self.num_entries):
            self.assertEqual(
                rb_find("struct drgn_test_rb_entry", self.root, "node", i, cmp),
                self.entry(i),
            )
        self.assertEqual(
            rb_find(
                "struct drgn_test_rb_entry", self.root, "node", self.num_entries, cmp
            ),
            NULL(self.prog, "struct drgn_test_rb_entry *"),
        )

    @staticmethod
    def cmp_entries(a, b):
        return a.value.value_() - b.value.value_()

    def test_validate_rbtree_success(self):
        for root, allow_equal in (
            (self.root, False),
            (self.empty_root, False),
            (self.prog["drgn_test_rbtree_with_equal"].address_of_(), True),
        ):
            validate_rbtree(
                "struct drgn_test_rb_entry", root, "node", self.cmp_entries, allow_equal
            )
            self.assertEqual(
                list(
                    validate_rbtree_inorder_for_each_entry(
                        "struct drgn_test_rb_entry",
                        root,
                        "node",
                        self.cmp_entries,
                        allow_equal,
                    )
                ),
                list(
                    rbtree_inorder_for_each_entry(
                        "struct drgn_test_rb_entry", root, "node"
                    )
                ),
            )

    def assert_validation_error(self, regex, name):
        self.assertRaisesRegex(
            ValidationError,
            regex,
            validate_rbtree,
            "struct drgn_test_rb_entry",
            self.prog[name].address_of_(),
            "node",
            self.cmp_entries,
            False,
        )
        self.assertRaisesRegex(
            ValidationError,
            regex,
            collections.deque,
            validate_rbtree_inorder_for_each_entry(
                "struct drgn_test_rb_entry",
                self.prog[name].address_of_(),
                "node",
                self.cmp_entries,
                False,
            ),
            0,
        )

    def test_validate_rbtree_has_equal(self):
        self.assert_validation_error("compares equal", "drgn_test_rbtree_with_equal")

    def test_validate_rbtree_out_of_order(self):
        self.assert_validation_error(
            "compares (greater|less) than", "drgn_test_rbtree_out_of_order"
        )

    def test_validate_rbtree_null_root_parent(self):
        self.assert_validation_error(
            "root node .* has parent", "drgn_test_rbtree_with_bad_root_parent"
        )

    def test_validate_rbtree_red_root(self):
        self.assert_validation_error(
            "root node .* is red", "drgn_test_rbtree_with_red_root"
        )

    def test_validate_rbtree_inconsistent_parents(self):
        self.assert_validation_error(
            "rb_parent", "drgn_test_rbtree_with_inconsistent_parents"
        )

    def test_validate_rbtree_red_violation(self):
        self.assert_validation_error(
            "red node .* has red child", "drgn_test_rbtree_with_red_violation"
        )

    def test_validate_rbtree_black_violation(self):
        self.assert_validation_error(
            "unequal black heights", "drgn_test_rbtree_with_black_violation"
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("max", type=int)
    args = parser.parse_args()

    for i in range(2, args.max + 1):
        print(
            f"#define PP_CAT{str(i) if i > 2 else ''}("
            + ", ".join(f"_{j}" for j in range(i))
            + f") PP_CAT_I{i}("
            + ", ".join(f"_{j}" for j in range(i))
            + ")"
        )
    print("/** @cond */")
    for i in range(2, args.max + 1):
        print(
            f"#define PP_CAT_I{i}("
            + ", ".join(f"_{j}" for j in range(i))
            + ") "
            + "##".join(f"_{j}" for j in range(i))
        )
    print("/** @endcond */")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later


import unittest.mock

from drgn.commands._crash.common import Cpuspec, parse_cpuspec
from drgn.helpers.linux.sched import idle_task
from tests import TestCase
from tests.linux_kernel import possible_cpus
from tests.linux_kernel.crash_commands import CrashCommandTestCase


class TestParseCpuspec(TestCase):
    def test_current(self):
        self.assertEqual(parse_cpuspec(""), Cpuspec(current=True))

    def test_all(self):
        self.assertEqual(parse_cpuspec("a"), Cpuspec(all=True))
        self.assertEqual(parse_cpuspec("all"), Cpuspec(all=True))

    def test_single(self):
        self.assertEqual(parse_cpuspec("0"), Cpuspec(explicit_cpus={0}))
        self.assertEqual(parse_cpuspec("13"), Cpuspec(explicit_cpus={13}))

    def test_multiple(self):
        self.assertEqual(parse_cpuspec("0,1,5"), Cpuspec(explicit_cpus={0, 1, 5}))

    def test_multiple_repeat(self):
        self.assertEqual(parse_cpuspec("1,0,1"), Cpuspec(explicit_cpus={0, 1}))

    def test_range(self):
        self.assertEqual(parse_cpuspec("0-2"), Cpuspec(explicit_cpus={0, 1, 2}))
        self.assertEqual(parse_cpuspec("10-12"), Cpuspec(explicit_cpus={10, 11, 12}))

    def test_range_single(self):
        self.assertEqual(parse_cpuspec("4-4"), Cpuspec(explicit_cpus={4}))

    def test_range_backwards(self):
        self.assertEqual(parse_cpuspec("4-3"), Cpuspec(explicit_cpus=set()))

    def test_bad_syntax(self):
        for spec in (
            ",",
            "0,",
            "0,,1",
            ",0",
            "0 0",
            "0-",
            "-1",
            "0,",
            "-",
            "+",
        ):
            with self.subTest(spec=spec):
                self.assertRaisesRegex(
                    ValueError, "invalid cpuspec", parse_cpuspec, spec
                )


class TestCpuspecCpus(CrashCommandTestCase):
    def setUp(self):
        patcher = unittest.mock.patch(
            "drgn.commands._crash.common.for_each_possible_cpu",
            side_effect=lambda prog: (0, 1, 2, 4, 5),
        )
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_current(self):
        cpu = max(possible_cpus())
        task = idle_task(self.prog, cpu)
        self.run_crash_command(f"set {hex(task)}")
        self.assertEqual(Cpuspec(current=True).cpus(self.prog), [cpu])

    def test_all(self):
        self.assertEqual(Cpuspec(all=True).cpus(self.prog), [0, 1, 2, 4, 5])

    def test_explicit(self):
        self.assertEqual(Cpuspec(explicit_cpus={0, 1, 4}).cpus(self.prog), [0, 1, 4])

    def test_invalid_cpu(self):
        self.assertRaisesRegex(
            ValueError,
            "invalid CPU",
            Cpuspec(explicit_cpus={3}).cpus,
            self.prog,
        )



# Copyright (c) 2026, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn.helpers.linux.completion import completion_done, completion_for_each_task
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestCompletion(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.completion = cls.prog["drgn_test_completion"].address_of_()
        cls.done_completion = cls.prog["drgn_test_done_completion"].address_of_()

    def test_completion_done(self):
        self.assertFalse(completion_done(self.completion))
        self.assertTrue(completion_done(self.done_completion))

    def test_completion_for_each_task(self):
        self.assertEqual(list(completion_for_each_task(self.done_completion)), [])
        self.assertEqual(
            list(completion_for_each_task(self.completion)),
            [self.prog["drgn_test_completion_kthread"]],
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

REGISTERS = [
    DrgnRegister(["r0", "a1"]),
    DrgnRegister(["r1", "a2"]),
    DrgnRegister(["r2", "a3"]),
    DrgnRegister(["r3", "a4"]),
    DrgnRegister(["r4", "v1"]),
    DrgnRegister(["r5", "v2"]),
    DrgnRegister(["r6", "v3"]),
    DrgnRegister(["r7", "v4"]),
    DrgnRegister(["r8", "v5"]),
    DrgnRegister(["r9", "v6", "sb"]),
    DrgnRegister(["r10", "v7", "sl"]),
    DrgnRegister(["r11", "v8", "fp"]),
    DrgnRegister(["r12", "ip"]),
    DrgnRegister(["r13", "sp"]),
    DrgnRegister(["r14", "lr"]),
    DrgnRegister(["r15", "pc"]),
]

REGISTER_LAYOUT = [
    DrgnRegisterLayout("r13", size=4, dwarf_number=13),
    DrgnRegisterLayout("r14", size=4, dwarf_number=14),
    # Callee-saved registers.
    *[DrgnRegisterLayout(f"r{i}", size=4, dwarf_number=i) for i in range(4, 12)],
    # Caller-saved registers.
    *[DrgnRegisterLayout(f"r{i}", size=4, dwarf_number=i) for i in range(4)],
    DrgnRegisterLayout("r12", size=4, dwarf_number=12),
    DrgnRegisterLayout("r15", size=4, dwarf_number=15),
    DrgnRegisterLayout("cpsr", size=4, dwarf_number=None),
]

STACK_POINTER_REGISTER = "r13"



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: MIT

"""
Extension to reference Linux kernel code on git.kernel.org.

The linux role links to a file in the Linux kernel code:

    :linux:`include/linux/list.h`

Or a specific line in a file:

    :linux:`include/linux/list.h:100`

By default, it links to Linus Torvald's master branch. This can be overriden
for the rest of the document with the linuxversion directive:

    :linuxversion: v6.7

Or for a specific link:

    :linux:`include/linux/list.h@v6.6`
    :linux:`include/linux/list.h:600@v6.6`

An explicit title can be given:

    :linux:`list_entry() <include/linux/list.h:600@v6.6>`

The linuxt role is the same as the linux role except that it formats the title
as inline text instead of inline code.
"""

import re
from typing import Any, Dict, List, Tuple

from docutils import nodes
from docutils.nodes import Node, system_message
import sphinx.application
import sphinx.util.docutils


class LinuxVersionDirective(sphinx.util.docutils.SphinxDirective):
    required_arguments = 1
    optional_arguments = 0

    def run(self) -> List[Node]:
        self.env.temp_data["linux_version"] = self.arguments[0]
        return []


class LinuxRole(sphinx.util.docutils.ReferenceRole):
    def __init__(self, code: bool) -> None:
        super().__init__()
        self._code = code

    def run(self) -> Tuple[List[Node], List[system_message]]:
        remainder, sep, head = self.target.rpartition("@")
        if not sep:
            remainder = head
            head = ""
        path, sep, line = remainder.rpartition(":")
        if not sep:
            path = line
            line = ""

        if not head:
            head = self.env.temp_data.get("linux_version", "master")

        if re.fullmatch(r"v[0-9]+\.[0-9]+.[0-9]+", head):
            base_url = (
                "https://git.kernel.org/pub/scm/linux/kernel/git/stable/linux.git/tree/"
            )
        else:
            base_url = "https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/"

        url_parts = [base_url, path]
        if head != "master":
            url_parts.append("?h=")
            url_parts.append(head)
        if line:
            url_parts.append("#n")
            url_parts.append(line)
        url = "".join(url_parts)

        if self.has_explicit_title:
            title = self.title
        else:
            title_parts = [path]
            if line:
                title_parts.append(":")
                title_parts.append(line)
            title = "".join(title_parts)
        if self._code:
            reference = nodes.reference("", "", internal=False, refuri=url)
            reference += nodes.literal(title, title)
        else:
            reference = nodes.reference(title, title, internal=False, refuri=url)

        return [reference], []


def setup(app: sphinx.application.Sphinx) -> Dict[str, Any]:
    app.add_directive("linuxversion", LinuxVersionDirective)
    app.add_role("linux", LinuxRole(True))
    app.add_role("linuxt", LinuxRole(False))
    return {"env_version": 1, "parallel_read_safe": True, "parallel_write_safe": True}



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import re
import sys

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generate PageFlag() helpers from include/linux/page-flags.h"
    )
    args = parser.parse_args()

    flags = {
        # PageUptodate() isn't defined with PAGEFLAG because it needs
        # additional memory barriers, but other than that it's the same.
        "Uptodate": "uptodate",
    }
    for match in re.finditer(
        r"\b(?:|__|TEST)PAGEFLAG\s*\(\s*(\w+)(?<!uname)\s*,\s*(\w+)\s*,\s*\w+\s*\)",
        sys.stdin.read(),
    ):
        if flags.setdefault(match.group(1), match.group(2)) != match.group(2):
            sys.exit(f"{match.group('uname')} has multiple lowercase names?")

    print("    # Generated by scripts/generate_page_flag_getters.py.")
    for uname, lname in sorted(flags.items()):
        print(f'    "Page{uname}",')
    print(")")
    print()
    for uname, lname in sorted(flags.items()):
        print(
            f'''
def Page{uname}(page: Object) -> bool:
    """
    Return whether the ``PG_{lname}`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_{lname}"]
    except KeyError:
        return False
    return bool(page_flags(page) & (1 << flag))
'''
        )
    print()
    print("# End generated by scripts/generate_page_flag_getters.py.")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import asyncio
import itertools
import logging
import os
from pathlib import Path
import re
import sys
from typing import (
    AsyncIterator,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import aiohttp
import uritemplate

from util import KernelVersion
from vmtest.asynciosubprocess import check_call, check_output
from vmtest.config import (
    ARCHITECTURES,
    KERNEL_FLAVORS,
    Architecture,
    Compiler,
    KernelFlavor,
    kconfig_localversion,
)
from vmtest.download import VMTEST_GITHUB_RELEASE, Downloader, available_kernel_releases
from vmtest.githubapi import AioGitHubApi
from vmtest.kbuild import KBuild, apply_patches

logger = logging.getLogger(__name__)

# [inclusive, exclusive) ranges of kernel versions to ignore when building
# latest releases of each version.
IGNORE_KERNEL_RANGES = (
    (KernelVersion("~"), KernelVersion("4.9")),
    (KernelVersion("4.10~"), KernelVersion("4.14")),
    (KernelVersion("4.15~"), KernelVersion("4.19")),
    (KernelVersion("4.20~"), KernelVersion("5.4")),
    (KernelVersion("5.5~"), KernelVersion("5.10")),
)

# Use the GitHub mirrors rather than the official kernel.org repositories since
# this script usually runs in GitHub Actions.
LINUX_GIT_URL = "https://github.com/torvalds/linux.git"
STABLE_LINUX_GIT_URL = "https://github.com/gregkh/linux.git"


async def get_latest_kernel_tags() -> List[str]:
    mainline_refs, stable_refs = await asyncio.gather(
        check_output("git", "ls-remote", "--tags", "--refs", LINUX_GIT_URL),
        check_output("git", "ls-remote", "--tags", "--refs", STABLE_LINUX_GIT_URL),
    )
    latest: Dict[str, KernelVersion] = {}
    for match in itertools.chain(
        re.finditer(
            r"^[a-f0-9]+\s+refs/tags/v([0-9]+\.[0-9]+)(-rc[0-9]+)?$",
            mainline_refs.decode(),
            re.M,
        ),
        re.finditer(
            r"^[a-f0-9]+\s+refs/tags/v([0-9]+\.[0-9]+)(\.[0-9]+)$",
            stable_refs.decode(),
            re.M,
        ),
    ):
        version = KernelVersion(match.group(1) + (match.group(2) or ""))
        for start_version, end_version in IGNORE_KERNEL_RANGES:
            if start_version <= version < end_version:
                break
        else:
            latest[match.group(1)] = max(version, latest.get(match.group(1), version))
    return ["v" + str(version) for version in sorted(latest.values(), reverse=True)]


async def fetch_kernel_tags(kernel_dir: Path, kernel_tags: Sequence[str]) -> None:
    if not kernel_dir.exists():
        logger.info("creating kernel repository in %s", kernel_dir)
        await check_call("git", "init", "-q", str(kernel_dir))

    mainline_tags = []
    stable_tags = []
    for tag in kernel_tags:
        if re.fullmatch(r"v[0-9]+\.[0-9]+\.[0-9]+", tag):
            stable_tags.append(tag)
        else:
            mainline_tags.append(tag)

    for name, url, tags in (
        ("mainline", LINUX_GIT_URL, mainline_tags),
        ("stable", STABLE_LINUX_GIT_URL, stable_tags),
    ):
        if tags:
            logger.info("fetching %s kernel tags: %s", name, ", ".join(tags))
            await check_call(
                "git",
                "-C",
                str(kernel_dir),
                "fetch",
                "--depth",
                "1",
                url,
                *(f"refs/tags/{tag}:refs/tags/{tag}" for tag in tags),
            )


async def build_kernels(
    kernel_dir: Path,
    build_dir: Path,
    kernel_revs: Sequence[
        Tuple[str, Sequence[Tuple[Architecture, Sequence[KernelFlavor]]]]
    ],
    compilers: Mapping[str, Compiler],
    keep_builds: bool,
) -> AsyncIterator[Path]:
    build_dir.mkdir(parents=True, exist_ok=True)
    for rev, arches in kernel_revs:
        logger.info("checking out %s in %s", rev, kernel_dir)
        await check_call(
            "git", "-C", str(kernel_dir), "checkout", "--force", "--quiet", rev
        )
        await check_call("git", "-C", str(kernel_dir), "clean", "-dqf")
        await apply_patches(kernel_dir)
        for arch, flavors in arches:
            env = {**os.environ, **compilers[arch.name].env()}
            for flavor in flavors:
                flavor_rev_build_dir = (
                    build_dir / f"build-{arch.name}-{flavor.name}-{rev}"
                )
                with open(
                    build_dir / f"build-{arch.name}-{flavor.name}-{rev}.log", "w"
                ) as build_log_file:
                    kbuild = KBuild(
                        kernel_dir,
                        flavor_rev_build_dir,
                        arch,
                        flavor,
                        env=env,
                        build_log_file=build_log_file,
                    )
                    await kbuild.build()
                    yield await kbuild.package("tar.zst", build_dir)
                    if not keep_builds:
                        logger.info("deleting %s", flavor_rev_build_dir)
                        # Shell out instead of using, e.g., shutil.rmtree(), to
                        # avoid blocking the main thread and the GIL.
                        await check_call("rm", "-rf", str(flavor_rev_build_dir))


class AssetUploadWork(NamedTuple):
    upload_url: str
    path: Union[str, bytes, Path]
    name: str
    content_type: str


async def asset_uploader(
    gh: AioGitHubApi,
    queue: "asyncio.Queue[Optional[AssetUploadWork]]",
) -> bool:
    success = True
    while True:
        work = await queue.get()
        if not work:
            queue.task_done()
            return success
        logger.info("uploading %s", work.name)
        try:
            with open(work.path, "rb") as f:
                await gh.upload(
                    uritemplate.expand(work.upload_url, name=work.name),
                    f,
                    work.content_type,
                )
        except Exception:
            logger.exception("uploading %s failed", work.name)
            success = False
        else:
            logger.info("uploaded %s", work.name)
        finally:
            queue.task_done()


async def main() -> None:
    logging.basicConfig(
        format="%(asctime)s:%(levelname)s:%(name)s:%(message)s", level=logging.INFO
    )

    parser = argparse.ArgumentParser(
        description="Build and upload drgn vmtest assets",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-K",
        "--latest-kernels",
        action="store_true",
        help="build and upload latest supported kernel releases",
    )
    parser.add_argument(
        "-a",
        "--architecture",
        dest="architectures",
        choices=["all", *sorted(ARCHITECTURES)],
        action="append",
        help="build architecture; may be given multiple times (default: all)",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-f",
        "--flavor",
        dest="flavors",
        choices=["all", *KERNEL_FLAVORS],
        action="append",
        help="build flavor; may be given multiple times (default: all)",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--no-build",
        dest="build",
        action="store_false",
        help="don't build or upload anything; just log what would be built",
    )
    parser.add_argument(
        "--no-upload",
        dest="upload",
        action="store_false",
        help="build but don't upload anything to GitHub",
    )
    parser.add_argument(
        "--kernel-directory",
        metavar="DIR",
        type=Path,
        help="kernel Git repository directory (created if needed)",
        default=".",
    )
    parser.add_argument(
        "--build-directory",
        metavar="DIR",
        type=Path,
        help="directory for build artifacts",
        default=".",
    )
    parser.add_argument(
        "--no-keep-builds",
        dest="keep_builds",
        action="store_false",
        help="delete kernel builds after packaging (default if uploading)",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--keep-builds",
        action="store_true",
        help="keep kernel builds after packaging (default if not uploading)",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--download-directory",
        metavar="DIR",
        type=Path,
        default="build/vmtest",
        help="directory to download assets to",
    )
    args = parser.parse_args()

    if not hasattr(args, "architectures") or "all" in args.architectures:
        args.architectures = list(ARCHITECTURES.values())
    else:
        args.architectures = [
            arch for arch in ARCHITECTURES.values() if arch.name in args.architectures
        ]
    if not hasattr(args, "flavors") or "all" in args.flavors:
        args.flavors = list(KERNEL_FLAVORS.values())
    else:
        args.flavors = [
            flavor for flavor in KERNEL_FLAVORS.values() if flavor.name in args.flavors
        ]

    if not args.build:
        args.upload = False
    if not hasattr(args, "keep_builds"):
        args.keep_builds = not args.upload

    async with aiohttp.ClientSession(trust_env=True) as session:
        GITHUB_TOKEN = os.getenv("GITHUB_TOKEN")
        if GITHUB_TOKEN is None and args.upload:
            sys.exit("GITHUB_TOKEN environment variable is not set")
        gh = AioGitHubApi(session, GITHUB_TOKEN)

        args.download_directory.mkdir(parents=True, exist_ok=True)
        github_release_coro = gh.get_release_by_tag(
            *VMTEST_GITHUB_RELEASE,
            cache=args.download_directory / "github_release.json",
        )
        if args.latest_kernels:
            github_release, latest_kernel_tags = await asyncio.gather(
                github_release_coro, get_latest_kernel_tags()
            )
        else:
            github_release = await github_release_coro

        kernel_releases = available_kernel_releases(github_release)
        logger.info("available kernel releases:")
        for arch in args.architectures:
            arch_kernel_releases = kernel_releases.get(arch.name, {})
            logger.info(
                "  %s: %s",
                arch.name,
                ", ".join(
                    sorted(arch_kernel_releases, key=KernelVersion, reverse=True)
                ),
            )

        to_build = []
        if args.latest_kernels:
            logger.info("latest kernel versions: %s", ", ".join(latest_kernel_tags))
            for tag in latest_kernel_tags:
                tag_arches_to_build = []

                match = re.fullmatch(r"v([0-9]+\.[0-9]+)(\.[0-9]+)?(-rc\d+)?", tag)
                assert match
                version = "".join(
                    [match.group(1), match.group(2) or ".0", match.group(3) or ""]
                )

                for arch in args.architectures:
                    if not arch.kernel_version_supported(version):
                        continue

                    arch_kernel_releases = kernel_releases.get(arch.name, {})
                    tag_arch_flavors_to_build = [
                        flavor
                        for flavor in args.flavors
                        if (version + kconfig_localversion(arch, flavor, version))
                        not in arch_kernel_releases
                    ]

                    if tag_arch_flavors_to_build:
                        tag_arches_to_build.append((arch, tag_arch_flavors_to_build))
                if tag_arches_to_build:
                    to_build.append((tag, tag_arches_to_build))

        if to_build:
            logger.info("kernel versions to build:")
            for tag, tag_arches_to_build in to_build:
                logger.info(
                    "  %s (%s)",
                    tag,
                    ", ".join(
                        [
                            f"{arch.name} [{', '.join([flavor.name for flavor in tag_arch_flavors_to_build])}]"
                            for arch, tag_arch_flavors_to_build in tag_arches_to_build
                        ]
                    ),
                )

            if args.build:
                downloader = Downloader(args.download_directory)
                compilers = {
                    arch.name: downloader.download_compiler(
                        downloader.resolve_compiler(arch)
                    )
                    for arch in {
                        arch.name: arch
                        for _, tag_arches_to_build in to_build
                        for arch, _ in tag_arches_to_build
                    }.values()
                }

                if args.upload:
                    upload_queue: "asyncio.Queue[Optional[AssetUploadWork]]" = (
                        asyncio.Queue()
                    )
                    uploader = asyncio.create_task(asset_uploader(gh, upload_queue))

                await fetch_kernel_tags(
                    args.kernel_directory, [tag for tag, _ in to_build]
                )

                async for kernel_package in build_kernels(
                    args.kernel_directory,
                    args.build_directory,
                    to_build,
                    compilers,
                    args.keep_builds,
                ):
                    if args.upload:
                        await upload_queue.put(
                            AssetUploadWork(
                                upload_url=github_release["upload_url"],
                                path=kernel_package,
                                name=kernel_package.name,
                                content_type="application/zstd",
                            )
                        )
                    else:
                        logger.info("would upload %s", kernel_package)

                if args.upload:
                    await upload_queue.put(None)
                    await upload_queue.join()
                    if not await uploader:
                        sys.exit("some uploads failed")
        else:
            logger.info("nothing to build")


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python3

import argparse
from pathlib import Path
import re
import sys
from typing import List

MACROS = (
    "NONE",
    "READ",
    "WRITE",
    "NRBITS",
    "TYPEBITS",
    "SIZEBITS",
    "DIRBITS",
)


SRCARCH_TO_MACHINE_NAMES = {
    "alpha": ["alpha"],
    "mips": ["mips", "mips64"],
    "parisc": ["parisc", "parisc64"],
    "powerpc": ["ppc", "ppc64"],
    "sparc": ["sparc", "sparc64"],
}


def add_ioc_macros(macros, machine_names: List[str], path: Path) -> None:
    contents = path.read_text()
    for name, definition in re.findall(
        rf"^\s*#\s*define\s+_IOC_({'|'.join(MACROS)})\s+(\S+)",
        contents,
        flags=re.M,
    ):
        match = re.fullmatch(r"([0-9]+)[UL]*", definition)
        for machine_name in machine_names:
            macros.setdefault(name, {})[machine_name] = int(match.group(1))


def main() -> None:
    argparse.ArgumentParser(
        description="Generate Python functions for ioctl numbers from the kernel source code"
    ).parse_args()

    macros = {}
    add_ioc_macros(macros, ["generic"], Path("include/uapi/asm-generic/ioctl.h"))
    for path in Path(".").glob("arch/*/include/uapi/asm/ioctl.h"):
        add_ioc_macros(macros, SRCARCH_TO_MACHINE_NAMES[path.parts[1]], path)

    for name in MACROS:
        by_machine_name = macros[name]
        generic = by_machine_name["generic"]
        values = [
            (arch, value) for arch, value in by_machine_name.items() if value != generic
        ]
        values.sort()
        if values:
            print(f"_IOC_{name} = {{")
            for arch, value in values:
                print(f"    {arch!r}: {value},")
            print(f"}}.get(NORMALIZED_MACHINE_NAME, {generic})")
        else:
            print(f"_IOC_{name} = {generic}")
        print()

    sys.stdout.write(
        """\
_IOC_NRSHIFT = 0
_IOC_TYPESHIFT = _IOC_NRSHIFT + _IOC_NRBITS
_IOC_SIZESHIFT = _IOC_TYPESHIFT + _IOC_TYPEBITS
_IOC_DIRSHIFT = _IOC_SIZESHIFT + _IOC_SIZEBITS


def _IOC(dir: int, type: int, nr: int, size: int) -> int:
    return (
        (dir << _IOC_DIRSHIFT)
        | (type << _IOC_TYPESHIFT)
        | (nr << _IOC_NRSHIFT)
        | (size << _IOC_SIZESHIFT)
    )


def _IO(type: int, nr: int) -> int:
    return _IOC(_IOC_NONE, type, nr, 0)


def _IOR(type: int, nr: int, size: int) -> int:
    return _IOC(_IOC_READ, type, nr, size)


def _IOW(type: int, nr: int, size: int) -> int:
    return _IOC(_IOC_WRITE, type, nr, size)


def _IOWR(type: int, nr: int, size: int) -> int:
    return _IOC(_IOC_READ | _IOC_WRITE, type, nr, size)
"""
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import re
import subprocess
import sys


def main() -> None:
    argparse.ArgumentParser(description="Generate elf.py from elf.h").parse_args()

    contents = subprocess.check_output(
        ["gcc", "-dD", "-E", "-"],
        input="#include <elf.h>\n",
        universal_newlines=True,
    )

    enums = {
        name: []
        for name in (
            "ET",
            "PT",
            "SHF",
            "SHN",
            "SHT",
            "STB",
            "STT",
            "STV",
        )
    }
    for match in re.finditer(
        r"^\s*#\s*define\s+(?P<enum>"
        + "|".join(enums)
        + r")_(?P<name>\w+)\s+(?:(?P<value>0x[0-9a-fA-F]+|[0-9]+)|(?:\(\s*1U?\s*<<\s*(?P<bitshift>[0-9]+)\s*\)))",
        contents,
        re.MULTILINE,
    ):
        enum = match.group("enum")
        name = match.group("name")
        if match.group("value"):
            value = int(match.group("value"), 0)
        else:
            value = 1 << int(match.group("bitshift"), 10)
        enums[enum].append((name, value))

    f = sys.stdout
    f.write(
        """\
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
# Generated by scripts/gen_elf_py.py.

import enum
"""
    )
    for type_name, constants in enums.items():
        assert constants
        enum_class = "IntFlag" if type_name == "SHF" else "IntEnum"
        f.write(f"\n\nclass {type_name}(enum.{enum_class}):\n")
        for name, value in constants:
            f.write(f"    {name} = 0x{value:X}\n")


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import unittest

from drgn import NULL
from drgn.helpers.linux.locking import (
    RwsemLocked,
    mutex_owner,
    rwsem_locked,
    rwsem_owner,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod
from util import KernelVersion


@skip_unless_have_test_kmod
class TestLocking(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.kthread = cls.prog["drgn_test_locking_kthread"].read_()

    @unittest.skipIf(
        KernelVersion(os.uname().release) < KernelVersion("4.10")
        and "SMP" not in os.uname().version,
        "mutex_owner() is not supported on Linux < 4.10 !SMP",
    )
    def test_mutex_owner(self):
        self.assertEqual(
            mutex_owner(self.prog["drgn_test_mutex_locked"].address_of_()),
            self.kthread,
        )
        self.assertEqual(
            mutex_owner(self.prog["drgn_test_mutex_unlocked"].address_of_()),
            NULL(self.prog, "struct task_struct *"),
        )

    @unittest.skipIf(
        KernelVersion(os.uname().release) < KernelVersion("5.3")
        and "SMP" not in os.uname().version,
        "rwsem_locked() is not supported on Linux < 5.3 !SMP",
    )
    def test_rwsem_locked(self):
        self.assertEqual(
            rwsem_locked(self.prog["drgn_test_rwsem_read_locked"].address_of_()),
            RwsemLocked.READ_LOCKED,
        )
        self.assertEqual(
            rwsem_locked(self.prog["drgn_test_rwsem_write_locked"].address_of_()),
            RwsemLocked.WRITE_LOCKED,
        )
        self.assertEqual(
            rwsem_locked(
                self.prog["drgn_test_rwsem_previously_read_locked"].address_of_()
            ),
            RwsemLocked.UNLOCKED,
        )
        self.assertEqual(
            rwsem_locked(
                self.prog["drgn_test_rwsem_previously_write_locked"].address_of_()
            ),
            RwsemLocked.UNLOCKED,
        )
        self.assertEqual(
            rwsem_locked(self.prog["drgn_test_rwsem_never_locked"].address_of_()),
            RwsemLocked.UNLOCKED,
        )
        self.assertEqual(
            rwsem_locked(self.prog["drgn_test_rwsem_writer_waiting"].address_of_()),
            RwsemLocked.READ_LOCKED,
        )

    @unittest.skipIf(
        KernelVersion(os.uname().release) < KernelVersion("5.3")
        and "SMP" not in os.uname().version,
        "rwsem_owner() is not supported on Linux < 5.3 !SMP",
    )
    def test_rwsem_owner(self):
        if KernelVersion(os.uname().release) >= KernelVersion("4.20"):
            self.assertEqual(
                rwsem_owner(self.prog["drgn_test_rwsem_read_locked"].address_of_()),
                self.kthread,
            )
            self.assertEqual(
                rwsem_owner(self.prog["drgn_test_rwsem_writer_waiting"].address_of_()),
                self.kthread,
            )
        else:
            self.assertEqual(
                rwsem_owner(self.prog["drgn_test_rwsem_read_locked"].address_of_()),
                NULL(self.prog, "struct task_struct *"),
            )
            self.assertEqual(
                rwsem_owner(self.prog["drgn_test_rwsem_writer_waiting"].address_of_()),
                NULL(self.prog, "struct task_struct *"),
            )
        self.assertEqual(
            rwsem_owner(self.prog["drgn_test_rwsem_write_locked"].address_of_()),
            self.kthread,
        )
        self.assertEqual(
            rwsem_owner(
                self.prog["drgn_test_rwsem_previously_write_locked"].address_of_()
            ),
            NULL(self.prog, "struct task_struct *"),
        )
        self.assertEqual(
            rwsem_owner(self.prog["drgn_test_rwsem_never_locked"].address_of_()),
            NULL(self.prog, "struct task_struct *"),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import re

from drgn.helpers.linux.vmstat import (
    global_numa_event_state,
    global_vm_event_state,
    nr_free_pages,
)
from tests.linux_kernel import LinuxKernelTestCase, meminfo_field_in_pages


class TestVmstat(LinuxKernelTestCase):
    def test_nr_free_pages(self):
        self.assertAlmostEqual(
            nr_free_pages(self.prog),
            meminfo_field_in_pages("MemFree"),
            delta=1024 * 1024 * 1024,
        )

    def test_global_numa_event_state(self):
        with open("/proc/vmstat", "r") as f:
            for line in f:
                match = re.match(r"numa_hit\s+([0-9]+)", line)
                if match:
                    expected = int(match.group(1))
                    break
            else:
                self.skipTest("kernel does not support NUMA statistics")
        self.assertAlmostEqual(
            global_numa_event_state(self.prog["NUMA_HIT"]), expected, delta=1024 * 1024
        )

    def test_global_vm_event_state(self):
        with open("/proc/vmstat", "r") as f:
            for line in f:
                match = re.match(r"pgmajfault\s+([0-9]+)", line)
                if match:
                    expected = int(match.group(1))
                    break
            else:
                self.skipTest("kernel does not support VM event statistics")
        self.assertAlmostEqual(
            global_vm_event_state(self.prog["PGMAJFAULT"]), expected, delta=1024 * 1024
        )
